# Copyright (C) 2012 Marie E. Rognes and David Ham
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# This demo illustrates use of finite element spaces defined over
# simplicies embedded in higher dimensions
from ufl import (Cell, FiniteElement, TestFunctions, TrialFunctions, div, dx,
                 inner)

# Define interval embedded in 3D:
domain = Cell("triangle", geometric_dimension=3)

# Define element over this domain
V = FiniteElement("RT", domain, 1)
Q = FiniteElement("DG", domain, 0)
element = V * Q

(u, p) = TrialFunctions(element)
(v, q) = TestFunctions(element)

a = (inner(u, v) + div(u) * q + div(v) * p) * dx



# Copyright (C) 2021 Matthew W. Scroggs and Chris Richardson
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Finite element interface."""

from __future__ import annotations

import typing
import warnings
from functools import lru_cache

import numpy

import basix
import basix.ufl_wrapper
import ufl


def convert_element(element: ufl.finiteelement.FiniteElementBase) -> basix.ufl_wrapper._BasixElementBase:
    """Convert and element to a FFCx element."""
    if isinstance(element, basix.ufl_wrapper._BasixElementBase):
        return element
    return create_element(element)


@lru_cache()
def create_element(element: ufl.finiteelement.FiniteElementBase) -> basix.ufl_wrapper._BasixElementBase:
    """Create an FFCx element from a UFL element.

    Args:
        element: A UFL finite element

    Returns:
        A Basix finite element
    """
    if isinstance(element, basix.ufl_wrapper._BasixElementBase):
        return element
    elif isinstance(element, ufl.VectorElement):
        return basix.ufl_wrapper.VectorElement(create_element(element.sub_elements()[0]), element.num_sub_elements())
    elif isinstance(element, ufl.TensorElement):
        if len(element.symmetry()) == 0:
            return basix.ufl_wrapper.TensorElement(create_element(element.sub_elements()[0]), element._value_shape)
        else:
            assert element.symmetry()[(1, 0)] == (0, 1)
            return basix.ufl_wrapper.TensorElement(create_element(
                element.sub_elements()[0]), element._value_shape, symmetric=True)
    elif isinstance(element, ufl.MixedElement):
        return basix.ufl_wrapper.MixedElement([create_element(e) for e in element.sub_elements()])
    elif isinstance(element, ufl.EnrichedElement):
        return basix.ufl_wrapper._create_enriched_element([create_element(e) for e in element._elements])
    elif element.family() == "Quadrature":
        return QuadratureElement(element.cell().cellname(), element.value_shape(), scheme=element.quadrature_scheme(),
                                 degree=element.degree())

    elif element.family() == "Real":
        return RealElement(element)
    else:
        return basix.ufl_wrapper.convert_ufl_element(element)


def basix_index(indices: typing.Tuple[int]) -> int:
    """Get the Basix index of a derivative."""
    return basix.index(*indices)


def create_quadrature(cellname, degree, rule) -> typing.Tuple[numpy.typing.NDArray[numpy.float64],
                                                              numpy.typing.NDArray[numpy.float64]]:
    """Create a quadrature rule."""
    if cellname == "vertex":
        return (numpy.ones((1, 0), dtype=numpy.float64), numpy.ones(1, dtype=numpy.float64))

    quadrature = basix.make_quadrature(
        basix.quadrature.string_to_type(rule), basix.cell.string_to_type(cellname), degree)

    # The quadrature degree from UFL can be very high for some
    # integrals.  Print warning if number of quadrature points
    # exceeds 100.
    num_points = quadrature[1].size
    if num_points >= 100:
        warnings.warn(
            f"Number of integration points per cell is: {num_points}. Consider using 'quadrature_degree' "
            "to reduce number.")
    return quadrature


def reference_cell_vertices(cellname: str) -> numpy.typing.NDArray[numpy.float64]:
    """Get the vertices of a reference cell."""
    return basix.geometry(basix.cell.string_to_type(cellname))


def map_facet_points(points: numpy.typing.NDArray[numpy.float64], facet: int,
                     cellname: str) -> numpy.typing.NDArray[numpy.float64]:
    """Map points from a reference facet to a physical facet."""
    geom = basix.geometry(basix.cell.string_to_type(cellname))
    facet_vertices = [geom[i] for i in basix.topology(basix.cell.string_to_type(cellname))[-2][facet]]
    return numpy.asarray([facet_vertices[0] + sum((i - facet_vertices[0]) * j for i, j in zip(facet_vertices[1:], p))
                          for p in points], dtype=numpy.float64)


class QuadratureElement(basix.ufl_wrapper._BasixElementBase):
    """A quadrature element."""

    _points: basix.ufl_wrapper._nda_f64
    _weights: basix.ufl_wrapper._nda_f64
    _entity_counts: typing.List[int]
    _cellname: str

    def __init__(
        self, cellname: str, value_shape: typing.Tuple[int, ...], scheme: typing.Optional[str] = None,
        degree: typing.Optional[int] = None, points: typing.Optional[basix.ufl_wrapper._nda_f64] = None,
        weights: typing.Optional[basix.ufl_wrapper._nda_f64] = None, mapname: str = "identity"
    ):
        """Initialise the element."""
        if scheme is not None:
            assert degree is not None
            assert points is None
            assert weights is None
            repr = f"QuadratureElement({cellname}, {scheme}, {degree})"
            self._points, self._weights = create_quadrature(cellname, degree, scheme)
        else:
            assert degree is None
            assert points is not None
            assert weights is not None
            self._points = points
            self._weights = weights
            repr = f"QuadratureElement({cellname}, {points}, {weights})"
            degree = len(points)

        self._cellname = cellname
        basix_cell = basix.cell.string_to_type(cellname)
        self._entity_counts = [len(i) for i in basix.topology(basix_cell)]

        super().__init__(repr, "quadrature element", cellname, value_shape, degree, mapname=mapname)

    def basix_sobolev_space(self):
        """Return the underlying Sobolev space."""
        return basix.sobolev_spaces.L2

    def __eq__(self, other) -> bool:
        """Check if two elements are equal."""
        return isinstance(other, QuadratureElement) and numpy.allclose(self._points, other._points)

    def __hash__(self) -> int:
        """Return a hash."""
        return super().__hash__()

    def tabulate(
        self, nderivs: int, points: basix.ufl_wrapper._nda_f64
    ) -> basix.ufl_wrapper._nda_f64:
        """Tabulate the basis functions of the element.

        Args:
            nderivs: Number of derivatives to tabulate.
            points: Points to tabulate at

        Returns:
            Tabulated basis functions
        """
        if nderivs > 0:
            raise ValueError("Cannot take derivatives of Quadrature element.")

        if points.shape != self._points.shape:
            raise ValueError("Mismatch of tabulation points and element points.")
        tables = numpy.asarray([numpy.eye(points.shape[0], points.shape[0])])
        return tables

    def get_component_element(self, flat_component: int) -> typing.Tuple[basix.ufl_wrapper._BasixElementBase, int, int]:
        """Get element that represents a component of the element, and the offset and stride of the component.

        Args:
            flat_component: The component

        Returns:
            component element, offset of the component, stride of the component
        """
        return self, 0, 1

    @property
    def ufcx_element_type(self) -> str:
        """Element type."""
        return "ufcx_quadrature_element"

    @property
    def dim(self) -> int:
        """Number of DOFs the element has."""
        return self._points.shape[0]

    @property
    def num_entity_dofs(self) -> typing.List[typing.List[int]]:
        """Number of DOFs associated with each entity."""
        dofs = []
        for d in self._entity_counts[:-1]:
            dofs += [[0] * d]

        dofs += [[self.dim]]
        return dofs

    @property
    def entity_dofs(self) -> typing.List[typing.List[typing.List[int]]]:
        """DOF numbers associated with each entity."""
        start_dof = 0
        entity_dofs = []
        for i in self.num_entity_dofs:
            dofs_list = []
            for j in i:
                dofs_list.append([start_dof + k for k in range(j)])
                start_dof += j
            entity_dofs.append(dofs_list)
        return entity_dofs

    @property
    def num_entity_closure_dofs(self) -> typing.List[typing.List[int]]:
        """Number of DOFs associated with the closure of each entity."""
        return self.num_entity_dofs

    @property
    def entity_closure_dofs(self) -> typing.List[typing.List[typing.List[int]]]:
        """DOF numbers associated with the closure of each entity."""
        return self.entity_dofs

    @property
    def num_global_support_dofs(self) -> int:
        """Get the number of global support DOFs."""
        return 0

    @property
    def reference_topology(self) -> typing.List[typing.List[typing.List[int]]]:
        """Topology of the reference element."""
        raise NotImplementedError()

    @property
    def reference_geometry(self) -> basix.ufl_wrapper._nda_f64:
        """Geometry of the reference element."""
        raise NotImplementedError()

    @property
    def family_name(self) -> str:
        """Family name of the element."""
        return "quadrature"

    @property
    def lagrange_variant(self) -> basix.LagrangeVariant:
        """Basix Lagrange variant used to initialise the element."""
        return None

    @property
    def dpc_variant(self) -> basix.DPCVariant:
        """Basix DPC variant used to initialise the element."""
        return None

    @property
    def element_family(self) -> basix.ElementFamily:
        """Basix element family used to initialise the element."""
        return None

    @property
    def cell_type(self) -> basix.CellType:
        """Basix cell type used to initialise the element."""
        return basix.cell.string_to_type(self._cellname)

    @property
    def discontinuous(self) -> bool:
        """True if the discontinuous version of the element is used."""
        return False

    @property
    def map_type(self) -> basix.MapType:
        """The Basix map type."""
        return basix.MapType.identity


class RealElement(basix.ufl_wrapper._BasixElementBase):
    """A real element."""

    _family_name: str
    _cellname: str
    _entity_counts: typing.List[int]

    def __init__(self, element: ufl.finiteelement.FiniteElementBase):
        """Initialise the element."""
        self._cellname = element.cell().cellname()
        self._family_name = element.family()
        tdim = element.cell().topological_dimension()

        self._entity_counts = []
        if tdim >= 1:
            self._entity_counts.append(element.cell().num_vertices())
        if tdim >= 2:
            self._entity_counts.append(element.cell().num_edges())
        if tdim >= 3:
            self._entity_counts.append(element.cell().num_facets())
        self._entity_counts.append(1)

        super().__init__(
            f"RealElement({element})", "real element", element.cell().cellname(), element.value_shape(),
            element.degree())

    def __eq__(self, other) -> bool:
        """Check if two elements are equal."""
        return isinstance(other, RealElement)

    def __hash__(self) -> int:
        """Return a hash."""
        return super().__hash__()

    def tabulate(
        self, nderivs: int, points: basix.ufl_wrapper._nda_f64
    ) -> basix.ufl_wrapper._nda_f64:
        """Tabulate the basis functions of the element.

        Args:
            nderivs: Number of derivatives to tabulate.
            points: Points to tabulate at

        Returns:
            Tabulated basis functions
        """
        out = numpy.zeros((nderivs + 1, len(points), 1))
        out[0, :] = 1.
        return out

    def get_component_element(self, flat_component: int) -> typing.Tuple[basix.ufl_wrapper._BasixElementBase, int, int]:
        """Get element that represents a component of the element, and the offset and stride of the component.

        Args:
            flat_component: The component

        Returns:
            component element, offset of the component, stride of the component
        """
        assert flat_component < self.value_size
        return self, 0, 1

    @property
    def ufcx_element_type(self) -> str:
        """Element type."""
        return "ufcx_real_element"

    @property
    def dim(self) -> int:
        """Number of DOFs the element has."""
        return 0

    @property
    def num_entity_dofs(self) -> typing.List[typing.List[int]]:
        """Number of DOFs associated with each entity."""
        dofs = []
        for d in self._entity_counts[:-1]:
            dofs += [[0] * d]

        dofs += [[self.dim]]
        return dofs

    @property
    def entity_dofs(self) -> typing.List[typing.List[typing.List[int]]]:
        """DOF numbers associated with each entity."""
        start_dof = 0
        entity_dofs = []
        for i in self.num_entity_dofs:
            dofs_list = []
            for j in i:
                dofs_list.append([start_dof + k for k in range(j)])
                start_dof += j
            entity_dofs.append(dofs_list)
        return entity_dofs

    @property
    def num_entity_closure_dofs(self) -> typing.List[typing.List[int]]:
        """Number of DOFs associated with the closure of each entity."""
        return self.num_entity_dofs

    @property
    def entity_closure_dofs(self) -> typing.List[typing.List[typing.List[int]]]:
        """DOF numbers associated with the closure of each entity."""
        return self.entity_dofs

    @property
    def num_global_support_dofs(self) -> int:
        """Get the number of global support DOFs."""
        return 1

    @property
    def reference_topology(self) -> typing.List[typing.List[typing.List[int]]]:
        """Topology of the reference element."""
        raise NotImplementedError()

    @property
    def reference_geometry(self) -> basix.ufl_wrapper._nda_f64:
        """Geometry of the reference element."""
        raise NotImplementedError()

    @property
    def family_name(self) -> str:
        """Family name of the element."""
        return self._family_name

    @property
    def lagrange_variant(self) -> basix.LagrangeVariant:
        """Basix Lagrange variant used to initialise the element."""
        return None

    @property
    def dpc_variant(self) -> basix.DPCVariant:
        """Basix DPC variant used to initialise the element."""
        return None

    @property
    def element_family(self) -> basix.ElementFamily:
        """Basix element family used to initialise the element."""
        return None

    @property
    def cell_type(self) -> basix.CellType:
        """Basix cell type used to initialise the element."""
        return basix.cell.string_to_type(self._cellname)

    @property
    def discontinuous(self) -> bool:
        """True if the discontinuous version of the element is used."""
        return False

    def basix_sobolev_space(self):
        """Return the underlying Sobolev space."""
        return basix.sobolev_spaces.Hinf

    @property
    def map_type(self) -> basix.MapType:
        """The Basix map type."""
        return basix.MapType.identity



#
# Author: Martin Sandve Alnes
# Date: 2008-12-22
#

# Modified by Garth N. Wells, 2009
from ufl import (Coefficient, Constant, FacetNormal, FiniteElement, Identity,
                 SpatialCoordinate, TensorElement, TestFunction, TrialFunction,
                 VectorElement, derivative, det, diff, dot, ds, dx, exp, grad,
                 inner, inv, tetrahedron, tr, variable)

# Cell and its properties
cell = tetrahedron
d = cell.geometric_dimension()
N = FacetNormal(cell)
x = SpatialCoordinate(cell)

# Elements
u_element = VectorElement("CG", cell, 2)
p_element = FiniteElement("CG", cell, 1)
A_element = TensorElement("CG", cell, 1)

# Test and trial functions
v = TestFunction(u_element)
w = TrialFunction(u_element)

# Displacement at current and two previous timesteps
u = Coefficient(u_element)
up = Coefficient(u_element)
upp = Coefficient(u_element)

# Time parameters
dt = Constant(cell)

# Fiber field
A = Coefficient(A_element)

# External forces
T = Coefficient(u_element)
p0 = Coefficient(p_element)

# Material parameters FIXME
rho = Constant(cell)
K = Constant(cell)
c00 = Constant(cell)
c11 = Constant(cell)
c22 = Constant(cell)

# Deformation gradient
I = Identity(d)
F = I + grad(u)
F = variable(F)
Finv = inv(F)
J = det(F)

# Left Cauchy-Green deformation tensor
B = F * F.T
I1_B = tr(B)
I2_B = (I1_B**2 - tr(B * B)) / 2
I3_B = J**2

# Right Cauchy-Green deformation tensor
C = F.T * F
I1_C = tr(C)
I2_C = (I1_C**2 - tr(C * C)) / 2
I3_C = J**2

# Green strain tensor
E = (C - I) / 2

# Mapping of strain in fiber directions
Ef = A * E * A.T

# Strain energy function W(Q(Ef))
Q = c00 * Ef[0, 0]**2 + c11 * Ef[1, 1]**2 + c22 * Ef[2, 2]**2  # FIXME: insert some simple law here
W = (K / 2) * (exp(Q) - 1)  # + p stuff

# First Piola-Kirchoff stress tensor
P = diff(W, F)

# Acceleration term discretized with finite differences
k = dt / rho
acc = (u - 2 * up + upp)

# Residual equation # FIXME: Can contain errors, not tested!
a_F = inner(acc, v) * dx \
    + k * inner(P, grad(v)) * dx \
    - k * dot(J * Finv * T, v) * ds(0) \
    - k * dot(J * Finv * p0 * N, v) * ds(1)

# Jacobi matrix of residual equation
a_J = derivative(a_F, u, w)

# Export forms
forms = [a_F, a_J]



# Copyright (C) 2021 Igor A. Baratta
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


import ufl
from ffcx.codegeneration.flop_count import count_flops


def create_form(degree):
    mesh = ufl.Mesh(ufl.VectorElement("Lagrange", "triangle", 1))
    element = ufl.FiniteElement("Lagrange", ufl.triangle, degree)
    V = ufl.FunctionSpace(mesh, element)

    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)

    return ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx + ufl.inner(u, v) * ufl.ds


def test_flops():
    k1, k2 = 2, 4
    a1 = create_form(k1)
    a2 = create_form(k2)

    dofs1 = (k1 + 1.) * (k1 + 2.) / 2.
    dofs2 = (k2 + 1.) * (k2 + 2.) / 2.

    flops_1 = count_flops(a1)
    assert len(flops_1) == 2

    flops_2 = count_flops(a2)
    assert len(flops_2) == 2

    r = sum(flops_2, 0.) / sum(flops_1, 0.)

    assert r > (dofs2**2 / dofs1**2)



# Copyright (C) 2019 Chris Richardson
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import sys

import ffcx.codegeneration.jit
import ufl


def test_cache_modes(compile_args):
    cell = ufl.triangle
    element = ufl.FiniteElement("Lagrange", cell, 1)
    u, v = ufl.TrialFunction(element), ufl.TestFunction(element)
    a = ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx
    forms = [a]

    # Load form from /tmp
    compiled_forms, module, code = ffcx.codegeneration.jit.compile_forms(forms, cffi_extra_compile_args=compile_args)
    tmpname = module.__name__
    tmpfile = module.__file__
    print(tmpname, tmpfile)
    del sys.modules[tmpname]

    # Load form from cache
    compiled_forms, module, code = ffcx.codegeneration.jit.compile_forms(
        forms, cache_dir="./compile-cache", cffi_extra_compile_args=compile_args)
    newname = module.__name__
    newfile = module.__file__
    print(newname, newfile)

    assert newname == tmpname
    assert newfile != tmpfile



import hashlib
import os

# Version of FFCx header files
__author__ = "FEniCS Project"
__license__ = "This code is released into the public domain"
__version__ = "2018.2.0.dev0"

# Get abspath on import, it can in some cases be a relative path w.r.t.
# curdir on startup
_include_path = os.path.dirname(os.path.abspath(__file__))


def get_include_path():
    """Return location of UFC header files."""
    return _include_path


def _compute_signature():
    # Compute signature of ufc header files
    h = hashlib.sha1()
    with open(os.path.join(get_include_path(), "ufcx.h")) as f:
        h.update(f.read().encode("utf-8"))
    return h.hexdigest()


_signature = _compute_signature()


def get_signature():
    """Return SHA-1 hash of the contents of ufcx.h.

    In this implementation, the value is computed on import.
    """
    return _signature



# Copyright (C) 2004-2020 Anders Logg, Garth N. Wells and Michal Habera
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Command-line interface to FFCx.

Parse command-line arguments and generate code from input UFL form files.
"""

import argparse
import cProfile
import logging
import pathlib
import re
import string

import ufl
from ffcx import __version__ as FFCX_VERSION
from ffcx import compiler, formatting
from ffcx.options import FFCX_DEFAULT_OPTIONS, get_options

logger = logging.getLogger("ffcx")

parser = argparse.ArgumentParser(
    description="FEniCS Form Compiler (FFCx, https://fenicsproject.org)")
parser.add_argument(
    "--version", action='version', version=f"%(prog)s (version {FFCX_VERSION})")
parser.add_argument("-o", "--output-directory", type=str, default=".", help="output directory")
parser.add_argument("--visualise", action="store_true", help="visualise the IR graph")
parser.add_argument("-p", "--profile", action='store_true', help="enable profiling")

# Add all options from FFCx option system
for opt_name, (opt_val, opt_desc) in FFCX_DEFAULT_OPTIONS.items():
    parser.add_argument(f"--{opt_name}",
                        type=type(opt_val), help=f"{opt_desc} (default={opt_val})")

parser.add_argument("ufl_file", nargs='+', help="UFL file(s) to be compiled")


def main(args=None):
    xargs = parser.parse_args(args)

    # Parse all other options
    priority_options = {k: v for k, v in xargs.__dict__.items() if v is not None}
    options = get_options(priority_options)

    # Call parser and compiler for each file
    for filename in xargs.ufl_file:
        file = pathlib.Path(filename)

        # Remove weird characters (file system allows more than the C
        # preprocessor)
        prefix = file.stem
        prefix = re.subn("[^{}]".format(string.ascii_letters + string.digits + "_"), "!", prefix)[0]
        prefix = re.subn("!+", "_", prefix)[0]

        # Turn on profiling
        if xargs.profile:
            pr = cProfile.Profile()
            pr.enable()

        # Load UFL file
        ufd = ufl.algorithms.load_ufl_file(filename)

        # Generate code
        code_h, code_c = compiler.compile_ufl_objects(
            ufd.forms + ufd.expressions + ufd.elements, ufd.object_names,
            prefix=prefix, options=options, visualise=xargs.visualise)

        # Write to file
        formatting.write_code(code_h, code_c, prefix, xargs.output_directory)

        # Turn off profiling and write status to file
        if xargs.profile:
            pr.disable()
            pfn = f"ffcx_{prefix}.profile"
            pr.dump_stats(pfn)

    return 0



# Copyright (C) 2013 Martin S. Alnaes
#
# A functional M involving a bunch of cell geometry quantities.
from ufl import (CellVolume, Circumradius, Coefficient, FacetArea, FacetNormal,
                 FiniteElement, SpatialCoordinate, ds, dx, tetrahedron)

cell = tetrahedron

V = FiniteElement("CG", cell, 1)
u = Coefficient(V)

# TODO: Add all geometry for all cell types to this and other demo files, need for regression test.
x = SpatialCoordinate(cell)
n = FacetNormal(cell)
vol = CellVolume(cell)
rad = Circumradius(cell)
area = FacetArea(cell)

M = u * (x[0] * vol * rad) * dx + u * (x[0] * vol * rad * area) * ds  # + u*area*avg(n[0]*x[0]*vol*rad)*dS



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Algorithms for working with multiindices."""

import ufl
from ufl.classes import ComponentTensor, FixedIndex, Index, Indexed
from ufl.permutation import compute_indices
from ufl.utils.indexflattening import flatten_multiindex, shape_to_strides


def map_indexed_arg_components(indexed):
    """Build a map from flattened components to subexpression.

    Builds integer list mapping between flattened components
    of indexed expression and its underlying tensor-valued subexpression.
    """
    assert isinstance(indexed, Indexed)

    # AKA indexed = tensor[multiindex]
    tensor, multiindex = indexed.ufl_operands

    # AKA e1 = e2[multiindex]
    # (this renaming is historical, but kept for consistency with all the variables *1,*2 below)
    e2 = tensor
    e1 = indexed

    # Get tensor and index shape
    sh1 = e1.ufl_shape
    sh2 = e2.ufl_shape
    fi1 = e1.ufl_free_indices
    fi2 = e2.ufl_free_indices
    fid1 = e1.ufl_index_dimensions
    fid2 = e2.ufl_index_dimensions

    # Compute regular and total shape
    tsh1 = sh1 + fid1
    tsh2 = sh2 + fid2
    # r1 = len(tsh1)
    r2 = len(tsh2)
    # str1 = shape_to_strides(tsh1)
    str2 = shape_to_strides(tsh2)
    assert not sh1
    assert sh2  # Must have shape to be indexed in the first place
    assert ufl.product(tsh1) <= ufl.product(tsh2)

    # Build map from fi2/fid2 position (-offset nmui) to fi1/fid1 position
    ind2_to_ind1_map = [None] * len(fi2)
    for k, i in enumerate(fi2):
        ind2_to_ind1_map[k] = fi1.index(i)

    # Build map from fi1/fid1 position to mi position
    nmui = len(multiindex)
    multiindex_to_ind1_map = [None] * nmui
    for k, i in enumerate(multiindex):
        if isinstance(i, Index):
            multiindex_to_ind1_map[k] = fi1.index(i.count())

    # Build map from flattened e1 component to flattened e2 component
    perm1 = compute_indices(tsh1)
    ni1 = ufl.product(tsh1)

    # Situation: e1 = e2[mi]
    d1 = [None] * ni1
    p2 = [None] * r2
    assert len(sh2) == nmui
    for k, i in enumerate(multiindex):
        if isinstance(i, FixedIndex):
            p2[k] = int(i)
    for c1, p1 in enumerate(perm1):
        for k, i in enumerate(multiindex):
            if isinstance(i, Index):
                p2[k] = p1[multiindex_to_ind1_map[k]]
        for k, i in enumerate(ind2_to_ind1_map):
            p2[nmui + k] = p1[i]
        c2 = flatten_multiindex(p2, str2)
        d1[c1] = c2

    # Consistency checks
    assert all(isinstance(x, int) for x in d1)
    assert len(set(d1)) == len(d1)
    return d1


def map_component_tensor_arg_components(tensor):
    """Build a map from flattened components to subexpression.

    Builds integer list mapping between flattended components
    of tensor and its underlying indexed subexpression.
    """
    assert isinstance(tensor, ComponentTensor)

    # AKA tensor = as_tensor(indexed, multiindex)
    indexed, multiindex = tensor.ufl_operands

    e1 = indexed
    e2 = tensor  # e2 = as_tensor(e1, multiindex)
    mi = [i for i in multiindex if isinstance(i, Index)]

    # Get tensor and index shapes
    sh1 = e1.ufl_shape  # (sh)ape of e1
    sh2 = e2.ufl_shape  # (sh)ape of e2
    fi1 = e1.ufl_free_indices  # (f)ree (i)ndices of e1
    fi2 = e2.ufl_free_indices  # ...
    fid1 = e1.ufl_index_dimensions  # (f)ree (i)ndex (d)imensions of e1
    fid2 = e2.ufl_index_dimensions  # ...

    # Compute total shape (tsh) of e1 and e2
    tsh1 = sh1 + fid1
    tsh2 = sh2 + fid2
    r1 = len(tsh1)  # 'total rank' or e1
    r2 = len(tsh2)  # ...
    str1 = shape_to_strides(tsh1)
    assert not sh1
    assert sh2
    assert len(mi) == len(multiindex)
    assert ufl.product(tsh1) == ufl.product(tsh2)
    assert fi1

    assert all(i in fi1 for i in fi2)

    nmui = len(multiindex)
    assert nmui == len(sh2)

    # Build map from fi2/fid2 position (-offset nmui) to fi1/fid1 position
    p2_to_p1_map = [None] * r2
    for k, i in enumerate(fi2):
        p2_to_p1_map[k + nmui] = fi1.index(i)

    # Build map from fi1/fid1 position to mi position
    for k, i in enumerate(mi):
        p2_to_p1_map[k] = fi1.index(mi[k].count())

    # Build map from flattened e1 component to flattened e2 component
    perm2 = compute_indices(tsh2)
    ni2 = ufl.product(tsh2)

    # Situation: e2 = as_tensor(e1, mi)
    d2 = [None] * ni2
    p1 = [None] * r1
    for c2, p2 in enumerate(perm2):
        for k2, k1 in enumerate(p2_to_p1_map):
            p1[k1] = p2[k2]
        c1 = flatten_multiindex(p1, str1)
        d2[c2] = c1

    # Consistency checks
    assert all(isinstance(x, int) for x in d2)
    assert len(set(d2)) == len(d2)
    return d2



# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Tools for indentation-aware code string stitching.

When formatting an AST into a string, it's better to collect lists of
snippets and then join them than adding the pieces continually, which
gives O(n^2) behaviour w.r.t. AST size n.

"""


class Indented(object):
    """Class to mark a collection of snippets for indentation.

    This way nested indentations can be handled by adding the prefix
    spaces only once to each line instead of splitting and indenting
    substrings repeatedly.

    """

    # Try to keep memory overhead low:
    __slots__ = ("body", )

    def __init__(self, body):
        # Body can be any valid snippet format
        self.body = body


def iter_indented_lines(snippets, level=0):
    """Iterate over indented string lines from a snippets data structure.

    The snippets object can be built recursively using the following types:

    - str: Split and yield as one line at a time indented to the appropriate level.

    - Indented: Yield the lines within this object indented by one level.

    - tuple,list: Yield lines from recursive application of this function to list items.

    """
    tabsize = 2
    indentation = ' ' * (tabsize * level)
    if isinstance(snippets, str):
        for line in snippets.split("\n"):
            yield indentation + line
    elif isinstance(snippets, Indented):
        for line in iter_indented_lines(snippets.body, level + 1):
            yield line
    elif isinstance(snippets, (tuple, list)):
        for part in snippets:
            for line in iter_indented_lines(part, level):
                yield line
    else:
        raise RuntimeError("Unexpected type %s:\n%s" % (type(snippets), str(snippets)))


def format_indented_lines(snippets, level=0):
    """Format recursive sequences of indented lines as one string."""
    return "\n".join(iter_indented_lines(snippets, level))



# Copyright (C) 2010-2011 Kristian B. Oelgaard
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# Illustration on how to use Conditional to define a source term
from ufl import (And, Constant, FiniteElement, Not, Or, SpatialCoordinate,
                 TestFunction, conditional, dx, ge, gt, le, lt, triangle)

element = FiniteElement("Lagrange", triangle, 2)

v = TestFunction(element)
g = Constant(triangle)

x = SpatialCoordinate(triangle)
c0 = conditional(le((x[0] - 0.33)**2 + (x[1] - 0.67)**2, 0.015), -1.0, 5.0)
c = conditional(le((x[0] - 0.33)**2 + (x[1] - 0.67)**2, 0.025), c0, 0.0)

t0 = And(ge(x[0], 0.55), le(x[0], 0.95))
t1 = Or(lt(x[1], 0.05), gt(x[1], 0.45))
t2 = And(t0, Not(t1))
t = conditional(And(ge(x[1] - x[0] - 0.05 + 0.55, 0.0), t2), -1.0, 0.0)

k = conditional(gt(1, 0), g, g + 1)

f = c + t + k

L = v * f * dx



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""FFCx/UFC specific symbol naming."""

import logging

import ufl.utils.derivativetuples

logger = logging.getLogger("ffcx")


# TODO: Get restriction postfix from somewhere central
def ufcx_restriction_postfix(restriction):
    if restriction == "+":
        res = "_0"
    elif restriction == "-":
        res = "_1"
    else:
        res = ""
    return res


def format_mt_name(basename, mt):
    """Format variable name for modified terminal."""
    access = str(basename)

    # Add averaged state to name
    if mt.averaged is not None:
        avg = f"_a{mt.averaged}"
        access += avg

    # Format restriction
    res = ufcx_restriction_postfix(mt.restriction).replace("_", "_r")
    access += res

    # Format global derivatives
    if mt.global_derivatives:
        assert basename == "J"
        der = f"_deriv_{''.join(map(str, mt.global_derivatives))}"
        access += der

    # Format local derivatives
    if mt.local_derivatives:
        # Convert "listing" derivative multindex into "counting" representation
        gdim = ufl.domain.extract_unique_domain(mt.terminal).geometric_dimension()
        ld_counting = ufl.utils.derivativetuples.derivative_listing_to_counts(mt.local_derivatives, gdim)
        der = f"_d{''.join(map(str, ld_counting))}"
        access += der

    # Add flattened component to name
    if mt.component:
        comp = f"_c{mt.flat_component}"
        access += comp

    return access


class FFCXBackendSymbols(object):
    """FFCx specific symbol definitions. Provides non-ufl symbols."""

    def __init__(self, language, coefficient_numbering, coefficient_offsets,
                 original_constant_offsets):
        self.L = language
        self.S = self.L.Symbol
        self.coefficient_numbering = coefficient_numbering
        self.coefficient_offsets = coefficient_offsets

        self.original_constant_offsets = original_constant_offsets

    def element_tensor(self):
        """Symbol for the element tensor itself."""
        return self.S("A")

    def entity(self, entitytype, restriction):
        """Entity index for lookup in element tables."""
        if entitytype == "cell":
            # Always 0 for cells (even with restriction)
            return self.L.LiteralInt(0)
        elif entitytype == "facet":
            postfix = "[0]"
            if restriction == "-":
                postfix = "[1]"
            return self.S("entity_local_index" + postfix)
        elif entitytype == "vertex":
            return self.S("entity_local_index[0]")
        else:
            logging.exception(f"Unknown entitytype {entitytype}")

    def argument_loop_index(self, iarg):
        """Loop index for argument #iarg."""
        indices = ["i", "j", "k", "l"]
        return self.S(indices[iarg])

    def coefficient_dof_sum_index(self):
        """Index for loops over coefficient dofs, assumed to never be used in two nested loops."""
        return self.S("ic")

    def quadrature_loop_index(self):
        """Reusing a single index name for all quadrature loops, assumed not to be nested."""
        return self.S("iq")

    def quadrature_permutation(self, index):
        """Quadrature permutation, as input to the function."""
        return self.S("quadrature_permutation")[index]

    def custom_weights_table(self):
        """Table for chunk of custom quadrature weights (including cell measure scaling)."""
        return self.S("weights_chunk")

    def custom_points_table(self):
        """Table for chunk of custom quadrature points (physical coordinates)."""
        return self.S("points_chunk")

    def weights_table(self, quadrature_rule):
        """Table of quadrature weights."""
        return self.S(f"weights_{quadrature_rule.id()}")

    def points_table(self, quadrature_rule):
        """Table of quadrature points (points on the reference integration entity)."""
        return self.S(f"points_{quadrature_rule.id()}")

    def x_component(self, mt):
        """Physical coordinate component."""
        return self.S(format_mt_name("x", mt))

    def J_component(self, mt):
        """Jacobian component."""
        # FIXME: Add domain number!
        return self.S(format_mt_name("J", mt))

    def domain_dof_access(self, dof, component, gdim, num_scalar_dofs, restriction):
        # FIXME: Add domain number or offset!
        offset = 0
        if restriction == "-":
            offset = num_scalar_dofs * 3
        vc = self.S("coordinate_dofs")
        return vc[3 * dof + component + offset]

    def domain_dofs_access(self, gdim, num_scalar_dofs, restriction):
        # FIXME: Add domain number or offset!
        return [
            self.domain_dof_access(dof, component, gdim, num_scalar_dofs, restriction)
            for dof in range(num_scalar_dofs) for component in range(gdim)
        ]

    def coefficient_dof_access(self, coefficient, dof_index):
        offset = self.coefficient_offsets[coefficient]
        w = self.S("w")
        return w[offset + dof_index]

    def coefficient_dof_access_blocked(self, coefficient: ufl.Coefficient, index,
                                       block_size, dof_offset):
        coeff_offset = self.coefficient_offsets[coefficient]
        w = self.S("w")
        _w = self.S(f"_w_{coeff_offset}_{dof_offset}")
        unit_stride_access = _w[index]
        original_access = w[coeff_offset + index * block_size + dof_offset]
        return unit_stride_access, original_access

    def coefficient_value(self, mt):
        """Symbol for variable holding value or derivative component of coefficient."""
        c = self.coefficient_numbering[mt.terminal]
        return self.S(format_mt_name("w%d" % (c, ), mt))

    def constant_index_access(self, constant, index):
        offset = self.original_constant_offsets[constant]
        c = self.S("c")

        return c[offset + index]

    def named_table(self, name):
        return self.S(name)

    def element_table(self, tabledata, entitytype, restriction):
        entity = self.entity(entitytype, restriction)

        if tabledata.is_uniform:
            entity = 0
        else:
            entity = self.entity(entitytype, restriction)

        if tabledata.is_piecewise:
            iq = 0
        else:
            iq = self.quadrature_loop_index()

        if tabledata.is_permuted:
            qp = self.quadrature_permutation(0)
            if restriction == "-":
                qp = self.quadrature_permutation(1)
        else:
            qp = 0

        # Return direct access to element table
        return self.named_table(tabledata.name)[qp][entity][iq]



# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numbers
import re

_subs = (
    # Remove 0s after e+ or e-
    (re.compile(r"e[\+]0*(.)"), r"e\1"),
    (re.compile(r"e[\-]0*(.)"), r"e-\1"),
)


def format_float(x, precision=None):
    """Format a float value according to given precision."""
    global _subs

    if precision:
        if isinstance(x, complex):
            s = "({:.{prec}}+I*{:.{prec}})".format(x.real, x.imag, prec=precision)
        elif isinstance(x, float):
            s = "{:.{prec}}".format(x, prec=precision)
        else:
            s = "{:.{prec}}".format(float(x), prec=precision)
    else:
        s = repr(float(x))
    for r, v in _subs:
        s = r.sub(v, s)
    return s


def format_int(x, precision=None):
    return str(x)


def format_value(value, precision=None):
    """Format a literal value as string.

    - float: Formatted according to current precision configuration.

    - int: Formatted as regular base 10 int literal.

    - str: Wrapped in "quotes".

    """
    if isinstance(value, numbers.Real):
        return format_float(float(value), precision=precision)
    elif isinstance(value, numbers.Integral):
        return format_int(int(value))
    elif isinstance(value, str):
        # FIXME: Is this ever used?
        assert '"' not in value
        return '"' + value + '"'
    elif hasattr(value, "ce_format"):
        return value.ce_format()
    else:
        raise RuntimeError("Unexpected type %s:\n%s" % (type(value),
                                                        str(value)))



# Copyright (C) 2009-2018 Anders Logg and Garth N. Wells
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Compiler stage 5: Code formatting.

This module implements the formatting of UFC code from a given
dictionary of generated C++ code for the body of each UFC function.

It relies on templates for UFC code available as part of the module
ufcx_utils.

"""

import logging
import os
import pprint
import textwrap

from ffcx import __version__ as FFCX_VERSION
from ffcx.codegeneration import __version__ as UFC_VERSION

logger = logging.getLogger("ffcx")

FORMAT_TEMPLATE = {
    "ufc comment":
    """\
// This code conforms with the UFC specification version {ufcx_version}
// and was automatically generated by FFCx version {ffcx_version}.
""",
    "dolfinx comment":
    """\
// This code conforms with the UFC specification version {ufcx_version}
// and was automatically generated by FFCx version {ffcx_version}.
//
""",
    "header_h":
    """
#pragma once

""",
    "header_c":
    """
""",
}

c_extern_pre = """
#ifdef __cplusplus
extern "C" {
#endif
"""

c_extern_post = """
#ifdef __cplusplus
}
#endif
"""


def format_code(code, options: dict):
    """Format given code in UFC format. Returns two strings with header and source file contents."""
    logger.info(79 * "*")
    logger.info("Compiler stage 5: Formatting code")
    logger.info(79 * "*")

    # Generate code for comment at top of file
    code_h_pre = _generate_comment(options) + "\n"
    code_c_pre = _generate_comment(options) + "\n"

    # Generate code for header
    code_h_pre += FORMAT_TEMPLATE["header_h"]
    code_c_pre += FORMAT_TEMPLATE["header_c"]

    # Generate includes and add to preamble
    includes_h, includes_c = _generate_includes(options)
    code_h_pre += includes_h
    code_c_pre += includes_c

    # Enclose header with 'extern "C"'
    code_h_pre += c_extern_pre
    code_h_post = c_extern_post

    code_h = ""
    code_c = ""

    for parts_code in code:
        code_h += "".join([c[0] for c in parts_code])
        code_c += "".join([c[1] for c in parts_code])

    # Add headers to body
    code_h = code_h_pre + code_h + code_h_post
    code_c = code_c_pre + code_c

    return code_h, code_c


def write_code(code_h, code_c, prefix, output_dir):
    _write_file(code_h, prefix, ".h", output_dir)
    _write_file(code_c, prefix, ".c", output_dir)


def _write_file(output, prefix, postfix, output_dir):
    """Write generated code to file."""
    filename = os.path.join(output_dir, prefix + postfix)
    with open(filename, "w") as hfile:
        hfile.write(output)


def _generate_comment(options):
    """Generate code for comment on top of file."""
    # Generate top level comment
    comment = FORMAT_TEMPLATE["ufc comment"].format(ffcx_version=FFCX_VERSION, ufcx_version=UFC_VERSION)

    # Add option information
    comment += "//\n"
    comment += "// This code was generated with the following options:\n"
    comment += "//\n"
    comment += textwrap.indent(pprint.pformat(options), "//  ")
    comment += "\n"

    return comment


def _generate_includes(options: dict):

    default_h_includes = [
        "#include <ufcx.h>",
    ]

    default_c_includes = [
        "#include <math.h>",  # This should really be set by the backend
        "#include <stdalign.h>",  # This should really be set by the backend
        "#include <stdlib.h>",  # This should really be set by the backend
        "#include <string.h>",  # This should really be set by the backend
        "#include <ufcx.h>"
    ]

    if "_Complex" in options["scalar_type"]:
        default_c_includes += ["#include <complex.h>"]

    s_h = set(default_h_includes)
    s_c = set(default_c_includes)

    includes_h = "\n".join(sorted(s_h)) + "\n" if s_h else ""
    includes_c = "\n".join(sorted(s_c)) + "\n" if s_c else ""

    return includes_h, includes_c



# Copyright (C) 2009-2020 Anders Logg, Martin Sandve Alnæs, Marie E. Rognes,
# Kristian B. Oelgaard, Matthew W. Scroggs, Chris Richardson, and others
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Compiler stage 2: Code representation.

Module computes intermediate representations of forms, elements and
dofmaps. For each UFC function, we extract the data needed for code
generation at a later stage.

The representation should conform strictly to the naming and order of
functions in UFC. Thus, for code generation of the function "foo", one
should only need to use the data stored in the intermediate
representation under the key "foo".
"""

import itertools
import logging
import numbers
import typing
import warnings

import numpy

import basix
import ufl
from ffcx import naming
from ffcx.analysis import UFLData
from ffcx.element_interface import convert_element
from ffcx.ir.integral import compute_integral_ir
from ffcx.ir.representationutils import (QuadratureRule,
                                         create_quadrature_points_and_weights)
from ufl.classes import Integral
from ufl.sorting import sorted_expr_sum

logger = logging.getLogger("ffcx")


class FormIR(typing.NamedTuple):
    id: int
    name: str
    signature: str
    rank: int
    num_coefficients: int
    num_constants: int
    name_from_uflfile: str
    function_spaces: typing.Dict[str, typing.Tuple[str, str, str, int, basix.CellType, basix.LagrangeVariant]]
    original_coefficient_position: typing.List[int]
    coefficient_names: typing.List[str]
    constant_names: typing.List[str]
    finite_elements: typing.List[str]
    dofmaps: typing.List[str]
    integral_names: typing.Dict[str, typing.List[str]]
    subdomain_ids: typing.Dict[str, typing.List[int]]


class CustomElementIR(typing.NamedTuple):
    cell_type: basix.CellType
    value_shape: typing.Tuple[int, ...]
    wcoeffs: numpy.typing.NDArray[numpy.float64]
    x: typing.List[typing.List[numpy.typing.NDArray[numpy.float64]]]
    M: typing.List[typing.List[numpy.typing.NDArray[numpy.float64]]]
    map_type: basix.MapType
    sobolev_space: basix.SobolevSpace
    interpolation_nderivs: int
    discontinuous: bool
    highest_complete_degree: int
    highest_degree: int


class ElementIR(typing.NamedTuple):
    id: int
    name: str
    signature: str
    cell_shape: str
    topological_dimension: int
    geometric_dimension: int
    space_dimension: int
    value_shape: typing.Tuple[int, ...]
    reference_value_shape: typing.Tuple[int, ...]
    degree: int
    family: str
    num_sub_elements: int
    block_size: int
    sub_elements: typing.List[str]
    element_type: str
    entity_dofs: typing.List[typing.List[typing.List[int]]]
    lagrange_variant: basix.LagrangeVariant
    dpc_variant: basix.DPCVariant
    basix_family: basix.ElementFamily
    basix_cell: basix.CellType
    discontinuous: bool
    custom_element: CustomElementIR


class DofMapIR(typing.NamedTuple):
    id: int
    name: str
    signature: str
    num_global_support_dofs: int
    num_element_support_dofs: int
    entity_dofs: typing.List[typing.List[typing.List[int]]]
    num_entity_dofs: typing.List[typing.List[int]]
    entity_closure_dofs: typing.List[typing.List[typing.List[int]]]
    num_entity_closure_dofs: typing.List[typing.List[int]]
    num_sub_dofmaps: int
    sub_dofmaps: typing.List[str]
    block_size: int


class IntegralIR(typing.NamedTuple):
    integral_type: str
    subdomain_id: typing.Union[str, typing.Tuple[int, ...], int]
    rank: int
    geometric_dimension: int
    topological_dimension: int
    entitytype: str
    num_facets: int
    num_vertices: int
    enabled_coefficients: typing.List[bool]
    element_dimensions: typing.Dict[ufl.FiniteElementBase, int]
    element_ids: typing.Dict[ufl.FiniteElementBase, int]
    tensor_shape: typing.List[int]
    coefficient_numbering: typing.Dict[ufl.Coefficient, int]
    coefficient_offsets: typing.Dict[ufl.Coefficient, int]
    original_constant_offsets: typing.Dict[ufl.Constant, int]
    options: dict
    cell_shape: str
    unique_tables: typing.Dict[str, numpy.typing.NDArray[numpy.float64]]
    unique_table_types: typing.Dict[str, str]
    integrand: typing.Dict[QuadratureRule, dict]
    name: str
    precision: int
    needs_facet_permutations: bool
    coordinate_element: str


class ExpressionIR(typing.NamedTuple):
    name: str
    element_dimensions: typing.Dict[ufl.FiniteElementBase, int]
    options: dict
    unique_tables: typing.Dict[str, numpy.typing.NDArray[numpy.float64]]
    unique_table_types: typing.Dict[str, str]
    integrand: typing.Dict[QuadratureRule, dict]
    coefficient_numbering: typing.Dict[ufl.Coefficient, int]
    coefficient_offsets: typing.Dict[ufl.Coefficient, int]
    integral_type: str
    entitytype: str
    tensor_shape: typing.List[int]
    expression_shape: typing.List[int]
    original_constant_offsets: typing.Dict[ufl.Constant, int]
    points: numpy.typing.NDArray[numpy.float64]
    coefficient_names: typing.List[str]
    constant_names: typing.List[str]
    needs_facet_permutations: bool
    function_spaces: typing.Dict[str, typing.Tuple[str, str, str, int, basix.CellType, basix.LagrangeVariant]]
    name_from_uflfile: str
    original_coefficient_positions: typing.List[int]


class DataIR(typing.NamedTuple):
    elements: typing.List[ElementIR]
    dofmaps: typing.List[DofMapIR]
    integrals: typing.List[IntegralIR]
    forms: typing.List[FormIR]
    expressions: typing.List[ExpressionIR]


def compute_ir(analysis: UFLData, object_names, prefix, options, visualise):
    """Compute intermediate representation."""
    logger.info(79 * "*")
    logger.info("Compiler stage 2: Computing intermediate representation of objects")
    logger.info(79 * "*")

    # Compute object names
    # NOTE: This is done here for performance reasons, because repeated calls
    # within each IR computation would be expensive due to UFL signature computations
    finite_element_names = {e: naming.finite_element_name(e, prefix) for e in analysis.unique_elements}
    dofmap_names = {e: naming.dofmap_name(e, prefix) for e in analysis.unique_elements}
    integral_names = {}
    form_names = {}
    for fd_index, fd in enumerate(analysis.form_data):
        form_names[fd_index] = naming.form_name(fd.original_form, fd_index, prefix)
        for itg_index, itg_data in enumerate(fd.integral_data):
            integral_names[(fd_index, itg_index)] = naming.integral_name(fd.original_form, itg_data.integral_type,
                                                                         fd_index, itg_data.subdomain_id, prefix)

    ir_elements = [
        _compute_element_ir(e, analysis.element_numbers, finite_element_names)
        for e in analysis.unique_elements
    ]

    ir_dofmaps = [
        _compute_dofmap_ir(e, analysis.element_numbers, dofmap_names)
        for e in analysis.unique_elements
    ]

    irs = [
        _compute_integral_ir(fd, i, analysis.element_numbers, integral_names, finite_element_names,
                             options, visualise)
        for (i, fd) in enumerate(analysis.form_data)
    ]
    ir_integrals = list(itertools.chain(*irs))

    ir_forms = [
        _compute_form_ir(fd, i, prefix, form_names, integral_names, analysis.element_numbers, finite_element_names,
                         dofmap_names, object_names)
        for (i, fd) in enumerate(analysis.form_data)
    ]

    ir_expressions = [_compute_expression_ir(expr, i, prefix, analysis, options, visualise, object_names,
                                             finite_element_names, dofmap_names)
                      for i, expr in enumerate(analysis.expressions)]

    return DataIR(elements=ir_elements, dofmaps=ir_dofmaps,
                  integrals=ir_integrals, forms=ir_forms,
                  expressions=ir_expressions)


def _compute_element_ir(element, element_numbers, finite_element_names):
    """Compute intermediate representation of element."""
    logger.info(f"Computing IR for element {element}")

    element = convert_element(element)

    # Create basix elements
    cell = element.cell()

    # Store id
    ir = {"id": element_numbers[element]}
    ir["name"] = finite_element_names[element]

    # Compute data for each function
    ir["signature"] = repr(element)
    ir["cell_shape"] = element.cell_type.name
    ir["topological_dimension"] = cell.topological_dimension()
    ir["geometric_dimension"] = cell.geometric_dimension()
    ir["space_dimension"] = element.dim + element.num_global_support_dofs
    ir["element_type"] = element.ufcx_element_type
    ir["lagrange_variant"] = element.lagrange_variant
    ir["dpc_variant"] = element.dpc_variant
    ir["basix_family"] = element.element_family
    ir["basix_cell"] = element.cell_type
    ir["discontinuous"] = element.discontinuous
    ir["degree"] = element.degree()
    ir["family"] = element.family_name
    ir["value_shape"] = element.value_shape()
    ir["reference_value_shape"] = element.reference_value_shape()

    ir["num_sub_elements"] = element.num_sub_elements()
    ir["sub_elements"] = [finite_element_names[e] for e in element.sub_elements()]

    ir["block_size"] = element.block_size
    if element.block_size > 1:
        element = element.sub_element

    ir["entity_dofs"] = element.entity_dofs

    if element.is_custom_element:
        ir["custom_element"] = _compute_custom_element_ir(element.element)
    else:
        ir["custom_element"] = None

    return ElementIR(**ir)


def _compute_custom_element_ir(basix_element: basix.finite_element.FiniteElement):
    """Compute intermediate representation of a custom Basix element."""
    ir = {}
    ir["cell_type"] = basix_element.cell_type
    ir["value_shape"] = basix_element.value_shape
    ir["wcoeffs"] = basix_element.wcoeffs
    ir["x"] = basix_element.x
    ir["M"] = basix_element.M
    ir["map_type"] = basix_element.map_type
    ir["sobolev_space"] = basix_element.sobolev_space
    ir["discontinuous"] = basix_element.discontinuous
    ir["interpolation_nderivs"] = basix_element.interpolation_nderivs
    ir["highest_complete_degree"] = basix_element.highest_complete_degree
    ir["highest_degree"] = basix_element.highest_degree

    return CustomElementIR(**ir)


def _compute_dofmap_ir(element, element_numbers, dofmap_names):
    """Compute intermediate representation of dofmap."""
    logger.info(f"Computing IR for dofmap of {element}")

    # Create basix elements
    element = convert_element(element)

    # Store id
    ir = {"id": element_numbers[element]}
    ir["name"] = dofmap_names[element]

    # Compute data for each function
    ir["signature"] = "FFCx dofmap for " + repr(element)
    ir["sub_dofmaps"] = [dofmap_names[e] for e in element.sub_elements()]
    ir["num_sub_dofmaps"] = element.num_sub_elements()

    ir["block_size"] = element.block_size
    if element.block_size > 1:
        element = element.sub_element

    # Precompute repeatedly used items
    for i in element.num_entity_dofs:
        # FIXME: this assumes the same number of DOFs on each entity of the same dim: this
        # assumption will not be true for prisms and pyramids
        if max(i) != min(i):
            raise RuntimeError("Elements with different numbers of DOFs on subentities of the same dimension"
                               " are not yet supported in FFCx.")

    # FIXME: This does not work for prisms and pyramids
    num_dofs_per_entity = [i[0] for i in element.num_entity_dofs]
    ir["num_entity_dofs"] = num_dofs_per_entity
    ir["entity_dofs"] = element.entity_dofs

    num_dofs_per_entity_closure = [i[0] for i in element.num_entity_closure_dofs]
    ir["num_entity_closure_dofs"] = num_dofs_per_entity_closure
    ir["entity_closure_dofs"] = element.entity_closure_dofs

    ir["num_global_support_dofs"] = element.num_global_support_dofs
    ir["num_element_support_dofs"] = element.dim

    return DofMapIR(**ir)


def _compute_integral_ir(form_data, form_index, element_numbers, integral_names,
                         finite_element_names, options, visualise):
    """Compute intermediate representation for form integrals."""
    _entity_types = {
        "cell": "cell",
        "exterior_facet": "facet",
        "interior_facet": "facet",
        "vertex": "vertex",
        "custom": "cell"
    }

    # Iterate over groups of integrals
    irs = []
    for itg_data_index, itg_data in enumerate(form_data.integral_data):

        logger.info(f"Computing IR for integral in integral group {itg_data_index}")

        # Compute representation
        entitytype = _entity_types[itg_data.integral_type]
        cell = itg_data.domain.ufl_cell()
        cellname = cell.cellname()
        tdim = cell.topological_dimension()
        assert all(tdim == itg.ufl_domain().topological_dimension() for itg in itg_data.integrals)

        ir = {
            "integral_type": itg_data.integral_type,
            "subdomain_id": itg_data.subdomain_id,
            "rank": form_data.rank,
            "geometric_dimension": form_data.geometric_dimension,
            "topological_dimension": tdim,
            "entitytype": entitytype,
            "num_facets": cell.num_facets(),
            "num_vertices": cell.num_vertices(),
            "enabled_coefficients": itg_data.enabled_coefficients,
            "cell_shape": cellname,
            "coordinate_element": finite_element_names[convert_element(itg_data.domain.ufl_coordinate_element())]
        }

        # Get element space dimensions
        unique_elements = element_numbers.keys()
        ir["element_dimensions"] = {element: element.dim + element.num_global_support_dofs
                                    for element in unique_elements}

        ir["element_ids"] = {
            element: i
            for i, element in enumerate(unique_elements)
        }

        # Create dimensions of primary indices, needed to reset the argument
        # 'A' given to tabulate_tensor() by the assembler.
        argument_dimensions = [
            ir["element_dimensions"][convert_element(element)] for element in form_data.argument_elements
        ]

        # Compute shape of element tensor
        if ir["integral_type"] == "interior_facet":
            ir["tensor_shape"] = [2 * dim for dim in argument_dimensions]
        else:
            ir["tensor_shape"] = argument_dimensions

        integral_type = itg_data.integral_type
        cell = itg_data.domain.ufl_cell()

        # Group integrands with the same quadrature rule
        grouped_integrands = {}
        for integral in itg_data.integrals:
            md = integral.metadata() or {}
            scheme = md["quadrature_rule"]

            if scheme == "custom":
                points = md["quadrature_points"]
                weights = md["quadrature_weights"]
            elif scheme == "vertex":
                # FIXME: Could this come from basix?

                # The vertex scheme, i.e., averaging the function value in the
                # vertices and multiplying with the simplex volume, is only of
                # order 1 and inferior to other generic schemes in terms of
                # error reduction. Equation systems generated with the vertex
                # scheme have some properties that other schemes lack, e.g., the
                # mass matrix is a simple diagonal matrix. This may be
                # prescribed in certain cases.
                degree = md["quadrature_degree"]
                if degree > 1:
                    warnings.warn(
                        "Explicitly selected vertex quadrature (degree 1), but requested degree is {}.".
                        format(degree))
                if cellname == "tetrahedron":
                    points, weights = (numpy.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0],
                                                    [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
                                       numpy.array([1.0 / 24.0, 1.0 / 24.0, 1.0 / 24.0, 1.0 / 24.0]))
                elif cellname == "triangle":
                    points, weights = (numpy.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]]),
                                       numpy.array([1.0 / 6.0, 1.0 / 6.0, 1.0 / 6.0]))
                elif cellname == "interval":
                    # Trapezoidal rule
                    return (numpy.array([[0.0], [1.0]]), numpy.array([1.0 / 2.0, 1.0 / 2.0]))
            else:
                degree = md["quadrature_degree"]
                points, weights = create_quadrature_points_and_weights(
                    integral_type, cell, degree, scheme)

            points = numpy.asarray(points)
            weights = numpy.asarray(weights)

            rule = QuadratureRule(points, weights)

            if rule not in grouped_integrands:
                grouped_integrands[rule] = []

            grouped_integrands[rule].append(integral.integrand())

        sorted_integrals = {}
        for rule, integrands in grouped_integrands.items():
            integrands_summed = sorted_expr_sum(integrands)

            integral_new = Integral(integrands_summed, itg_data.integral_type, itg_data.domain,
                                    itg_data.subdomain_id, {}, None)
            sorted_integrals[rule] = integral_new

        # TODO: See if coefficient_numbering can be removed
        # Build coefficient numbering for UFC interface here, to avoid
        # renumbering in UFL and application of replace mapping
        coefficient_numbering = {}
        for i, f in enumerate(form_data.reduced_coefficients):
            coefficient_numbering[f] = i

        # Add coefficient numbering to IR
        ir["coefficient_numbering"] = coefficient_numbering

        index_to_coeff = sorted([(v, k) for k, v in coefficient_numbering.items()])
        offsets = {}
        width = 2 if integral_type in ("interior_facet") else 1
        _offset = 0
        for k, el in zip(index_to_coeff, form_data.coefficient_elements):
            offsets[k[1]] = _offset
            _offset += width * ir["element_dimensions"][convert_element(el)]

        # Copy offsets also into IR
        ir["coefficient_offsets"] = offsets

        # Build offsets for Constants
        original_constant_offsets = {}
        _offset = 0
        for constant in form_data.original_form.constants():
            original_constant_offsets[constant] = _offset
            _offset += numpy.product(constant.ufl_shape, dtype=int)

        ir["original_constant_offsets"] = original_constant_offsets
        ir["precision"] = itg_data.metadata["precision"]

        # Create map from number of quadrature points -> integrand
        integrands = {rule: integral.integrand() for rule, integral in sorted_integrals.items()}

        # Build more specific intermediate representation
        integral_ir = compute_integral_ir(itg_data.domain.ufl_cell(), itg_data.integral_type,
                                          ir["entitytype"], integrands, ir["tensor_shape"],
                                          options, visualise)

        ir.update(integral_ir)

        # Fetch name
        ir["name"] = integral_names[(form_index, itg_data_index)]

        irs.append(IntegralIR(**ir))

    return irs


def _compute_form_ir(form_data, form_id, prefix, form_names, integral_names, element_numbers, finite_element_names,
                     dofmap_names, object_names) -> FormIR:
    """Compute intermediate representation of form."""
    logger.info(f"Computing IR for form {form_id}")

    # Store id
    ir = {"id": form_id}

    # Compute common data
    ir["name"] = form_names[form_id]

    ir["signature"] = form_data.original_form.signature()

    ir["rank"] = len(form_data.original_form.arguments())
    ir["num_coefficients"] = len(form_data.reduced_coefficients)
    ir["num_constants"] = len(form_data.original_form.constants())

    ir["coefficient_names"] = [object_names.get(id(obj), f"w{j}")
                               for j, obj in enumerate(form_data.reduced_coefficients)]

    ir["constant_names"] = [object_names.get(id(obj), f"c{j}")
                            for j, obj in enumerate(form_data.original_form.constants())]

    ir["original_coefficient_position"] = form_data.original_coefficient_positions

    ir["finite_elements"] = [
        finite_element_names[convert_element(e)]
        for e in form_data.argument_elements + form_data.coefficient_elements
    ]

    ir["dofmaps"] = [
        dofmap_names[convert_element(e)] for e in form_data.argument_elements + form_data.coefficient_elements
    ]

    fs = {}
    for function in form_data.original_form.arguments() + tuple(form_data.reduced_coefficients):
        name = object_names.get(id(function), str(function))
        el = convert_element(convert_element(function.ufl_function_space().ufl_element()))
        cmap = function.ufl_function_space().ufl_domain().ufl_coordinate_element()
        # Default point spacing for CoordinateElement is equispaced
        if not isinstance(cmap, basix.ufl_wrapper._BasixElementBase) and cmap.variant() is None:
            cmap._sub_element._variant = "equispaced"
        cmap = convert_element(cmap)
        family = cmap.family()
        degree = cmap.degree()
        fs[name] = (finite_element_names[el], dofmap_names[el], family, degree,
                    cmap.cell_type, cmap.lagrange_variant)

    form_name = object_names.get(id(form_data.original_form), form_id)

    ir["function_spaces"] = fs
    ir["name_from_uflfile"] = f"form_{prefix}_{form_name}"

    # Store names of integrals and subdomain_ids for this form, grouped
    # by integral types Since form points to all integrals it contains,
    # it has to know their names for codegen phase
    ir["integral_names"] = {}
    ir["subdomain_ids"] = {}
    ufcx_integral_types = ("cell", "exterior_facet", "interior_facet")
    for integral_type in ufcx_integral_types:
        ir["subdomain_ids"][integral_type] = []
        ir["integral_names"][integral_type] = []

        for itg_index, itg_data in enumerate(form_data.integral_data):
            if (itg_data.integral_type == integral_type):
                if itg_data.subdomain_id == "otherwise":
                    # UFL is using "otherwise" for default integrals
                    # (over whole mesh) but FFCx needs integers, so
                    # otherwise = -1
                    if len(ir["subdomain_ids"][integral_type]) > 0 and ir["subdomain_ids"][integral_type][0] == -1:
                        raise ValueError("Only one default ('otherwise') integral allowed.")

                    # Put default integral as first
                    ir["subdomain_ids"][integral_type] = [-1] + ir["subdomain_ids"][integral_type]
                    ir["integral_names"][integral_type] = [
                        integral_names[(form_id, itg_index)]] + ir["integral_names"][integral_type]
                elif itg_data.subdomain_id < 0:
                    raise ValueError("Integral subdomain ID must be non-negative.")
                else:
                    assert isinstance(itg_data.subdomain_id, numbers.Integral)
                    ir["subdomain_ids"][integral_type] += [itg_data.subdomain_id]
                    ir["integral_names"][integral_type] += [integral_names[(form_id, itg_index)]]

    return FormIR(**ir)


def _compute_expression_ir(expression, index, prefix, analysis, options, visualise, object_names,
                           finite_element_names, dofmap_names):
    """Compute intermediate representation of expression."""
    logger.info(f"Computing IR for expression {index}")

    # Compute representation
    ir = {}

    original_expression = (expression[2], expression[1])

    ir["name"] = naming.expression_name(original_expression, prefix)

    original_expression = expression[2]
    points = expression[1]
    expression = expression[0]

    try:
        cell = ufl.domain.extract_unique_domain(expression).ufl_cell()
    except AttributeError:
        # This case corresponds to a spatially constant expression
        # without any dependencies
        cell = None

    # Prepare dimensions of all unique element in expression, including
    # elements for arguments, coefficients and coordinate mappings
    ir["element_dimensions"] = {element: element.dim + element.num_global_support_dofs
                                for element in analysis.unique_elements}

    # Extract dimensions for elements of arguments only
    arguments = ufl.algorithms.extract_arguments(expression)
    argument_elements = tuple(convert_element(f.ufl_function_space().ufl_element()) for f in arguments)
    argument_dimensions = [ir["element_dimensions"][element] for element in argument_elements]

    tensor_shape = argument_dimensions
    ir["tensor_shape"] = tensor_shape

    ir["expression_shape"] = list(expression.ufl_shape)

    coefficients = ufl.algorithms.extract_coefficients(expression)
    coefficient_numbering = {}
    for i, coeff in enumerate(coefficients):
        coefficient_numbering[coeff] = i

    # Add coefficient numbering to IR
    ir["coefficient_numbering"] = coefficient_numbering

    original_coefficient_positions = []
    original_coefficients = ufl.algorithms.extract_coefficients(original_expression)
    for coeff in coefficients:
        original_coefficient_positions.append(original_coefficients.index(coeff))

    ir["coefficient_names"] = [object_names.get(id(obj), f"w{j}")
                               for j, obj in enumerate(coefficients)]

    ir["constant_names"] = [object_names.get(id(obj), f"c{j}")
                            for j, obj in enumerate(ufl.algorithms.analysis.extract_constants(expression))]

    fs = {}
    for function in tuple(original_coefficients) + tuple(arguments):
        name = object_names.get(id(function), str(function))
        el = convert_element(function.ufl_function_space().ufl_element())
        cmap = convert_element(function.ufl_function_space().ufl_domain().ufl_coordinate_element())
        family = cmap.family()
        degree = cmap.degree()
        fs[name] = (finite_element_names[el], dofmap_names[el], family, degree)

    expression_name = object_names.get(id(original_expression), index)

    ir["function_spaces"] = fs
    ir["name_from_uflfile"] = f"expression_{prefix}_{expression_name}"

    if len(argument_elements) > 1:
        raise RuntimeError("Expression with more than one Argument not implemented.")

    ir["original_coefficient_positions"] = original_coefficient_positions

    coefficient_elements = tuple(convert_element(f.ufl_element()) for f in coefficients)

    offsets = {}
    _offset = 0
    for i, el in enumerate(coefficient_elements):
        offsets[coefficients[i]] = _offset
        _offset += ir["element_dimensions"][convert_element(el)]

    # Copy offsets also into IR
    ir["coefficient_offsets"] = offsets

    ir["integral_type"] = "expression"
    ir["entitytype"] = "cell"

    # Build offsets for Constants
    original_constant_offsets = {}
    _offset = 0
    for constant in ufl.algorithms.analysis.extract_constants(expression):
        original_constant_offsets[constant] = _offset
        _offset += numpy.product(constant.ufl_shape, dtype=int)

    ir["original_constant_offsets"] = original_constant_offsets

    ir["points"] = points

    weights = numpy.array([1.0] * points.shape[0])
    rule = QuadratureRule(points, weights)
    integrands = {rule: expression}

    if cell is None:
        assert len(ir["original_coefficient_positions"]) == 0 and len(ir["original_constant_offsets"]) == 0

    expression_ir = compute_integral_ir(cell, ir["integral_type"], ir["entitytype"], integrands, tensor_shape,
                                        options, visualise)

    ir.update(expression_ir)

    return ExpressionIR(**ir)



# Code generation format strings for UFC (Unified Form-assembly Code)
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2018

declaration = """
extern ufcx_integral {factory_name};
"""

factory = """
// Code for integral {factory_name}

void tabulate_tensor_{factory_name}({scalar_type}* restrict A,
                                    const {scalar_type}* restrict w,
                                    const {scalar_type}* restrict c,
                                    const {geom_type}* restrict coordinate_dofs,
                                    const int* restrict entity_local_index,
                                    const uint8_t* restrict quadrature_permutation)
{{
{tabulate_tensor}
}}

{enabled_coefficients_init}

ufcx_integral {factory_name} =
{{
  .enabled_coefficients = {enabled_coefficients},
  .tabulate_tensor_{np_scalar_type} = tabulate_tensor_{factory_name},
  .needs_facet_permutations = {needs_facet_permutations},
  .coordinate_element = {coordinate_element},
}};

// End of code for integral {factory_name}
"""



# Copyright (C) 2007-2020 Anders Logg and Michal Habera
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Main interface for compilation of forms.

Breaks the compilation into several sequential stages.
The output of each stage is the input of the next stage.

Compiler stages
---------------

0. Language, parsing

   - Input:  Python code or .ufl file
   - Output: UFL form

   This stage consists of parsing and expressing a form in the UFL form
   language. This stage is handled by UFL.

1. Analysis

   - Input:  UFL form
   - Output: Preprocessed UFL form and FormData (metadata)

   This stage preprocesses the UFL form and extracts form metadata. It
   may also perform simplifications on the form.

2. Code representation

   - Input:  Preprocessed UFL form and FormData (metadata)
   - Output: Intermediate Representation (IR)

   This stage examines the input and generates all data needed for code
   generation. This includes generation of finite element basis
   functions, extraction of data for mapping of degrees of freedom and
   possible precomputation of integrals. Most of the complexity of
   compilation is handled in this stage.

   The IR is stored as a dictionary, mapping names of UFC functions to
   data needed for generation of the corresponding code.

3. Code generation

   - Input:  Intermediate Representation (IR)
   - Output: C code

   This stage examines the IR and generates the actual C code for the
   body of each UFC function.

   The code is stored as a dictionary, mapping names of UFC functions to
   strings containing the C code of the body of each function.

4. Code formatting

   - Input:  C code
   - Output: C code files

   This stage examines the generated C++ code and formats it according
   to the UFC format, generating as output one or more .h/.c files
   conforming to the UFC format.

"""

import logging
import typing
from time import time

from ffcx.analysis import analyze_ufl_objects
from ffcx.codegeneration.codegeneration import generate_code
from ffcx.formatting import format_code
from ffcx.ir.representation import compute_ir

logger = logging.getLogger("ffcx")


def _print_timing(stage: int, timing: float):
    logger.info(f"Compiler stage {stage} finished in {timing:.4f} seconds.")


def compile_ufl_objects(ufl_objects: typing.List[typing.Any],
                        object_names: typing.Dict = {},
                        prefix: typing.Optional[str] = None,
                        options: typing.Dict = {},
                        visualise: bool = False):
    """Generate UFC code for a given UFL objects.

    Options
    ----------
    @param ufl_objects:
        Objects to be compiled. Accepts elements, forms, integrals or coordinate mappings.

    """
    # Stage 1: analysis
    cpu_time = time()
    analysis = analyze_ufl_objects(ufl_objects, options)
    _print_timing(1, time() - cpu_time)

    # Stage 2: intermediate representation
    cpu_time = time()
    ir = compute_ir(analysis, object_names, prefix, options, visualise)
    _print_timing(2, time() - cpu_time)

    # Stage 3: code generation
    cpu_time = time()
    code = generate_code(ir, options)
    _print_timing(3, time() - cpu_time)

    # Stage 4: format code
    cpu_time = time()
    code_h, code_c = format_code(code, options)
    _print_timing(4, time() - cpu_time)

    return code_h, code_c



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Tools for C/C++ expression formatting."""

import logging

import ufl

logger = logging.getLogger("ffcx")

# Table of handled math functions for different scalar types

math_table = {'double': {'sqrt': 'sqrt',
                         'abs': 'fabs',
                         'cos': 'cos',
                         'sin': 'sin',
                         'tan': 'tan',
                         'acos': 'acos',
                         'asin': 'asin',
                         'atan': 'atan',
                         'cosh': 'cosh',
                         'sinh': 'sinh',
                         'tanh': 'tanh',
                         'acosh': 'acosh',
                         'asinh': 'asinh',
                         'atanh': 'atanh',
                         'power': 'pow',
                         'exp': 'exp',
                         'ln': 'log',
                         'erf': 'erf',
                         'atan_2': 'atan2',
                         'min_value': 'fmin',
                         'max_value': 'fmax'},

              'float': {'sqrt': 'sqrtf',
                        'abs': 'fabsf',
                        'cos': 'cosf',
                        'sin': 'sinf',
                        'tan': 'tanf',
                        'acos': 'acosf',
                        'asin': 'asinf',
                        'atan': 'atanf',
                        'cosh': 'coshf',
                        'sinh': 'sinhf',
                        'tanh': 'tanhf',
                        'acosh': 'acoshf',
                        'asinh': 'asinhf',
                        'atanh': 'atanhf',
                        'power': 'powf',
                        'exp': 'expf',
                        'ln': 'logf',
                        'erf': 'erff',
                        'atan_2': 'atan2f',
                        'min_value': 'fminf',
                        'max_value': 'fmaxf'},

              'long double': {'sqrt': 'sqrtl',
                              'abs': 'fabsl',
                              'cos': 'cosl',
                              'sin': 'sinl',
                              'tan': 'tanl',
                              'acos': 'acosl',
                              'asin': 'asinl',
                              'atan': 'atanl',
                              'cosh': 'coshl',
                              'sinh': 'sinhl',
                              'tanh': 'tanhl',
                              'acosh': 'acoshl',
                              'asinh': 'asinhl',
                              'atanh': 'atanhl',
                              'power': 'powl',
                              'exp': 'expl',
                              'ln': 'logl',
                              'erf': 'erfl',
                              'atan_2': 'atan2l',
                              'min_value': 'fminl',
                              'max_value': 'fmaxl'},

              'double _Complex': {'sqrt': 'csqrt',
                                  'abs': 'cabs',
                                  'cos': 'ccos',
                                  'sin': 'csin',
                                  'tan': 'ctan',
                                  'acos': 'cacos',
                                  'asin': 'casin',
                                  'atan': 'catan',
                                  'cosh': 'ccosh',
                                  'sinh': 'csinh',
                                  'tanh': 'ctanh',
                                  'acosh': 'cacosh',
                                  'asinh': 'casinh',
                                  'atanh': 'catanh',
                                  'power': 'cpow',
                                  'exp': 'cexp',
                                  'ln': 'clog',
                                  'real': 'creal',
                                  'imag': 'cimag',
                                  'conj': 'conj',
                                  'max_value': 'fmax',
                                  'min_value': 'fmin'},

              'float _Complex': {'sqrt': 'csqrtf',
                                 'abs': 'cabsf',
                                 'cos': 'ccosf',
                                 'sin': 'csinf',
                                 'tan': 'ctanf',
                                 'acos': 'cacosf',
                                 'asin': 'casinf',
                                 'atan': 'catanf',
                                 'cosh': 'ccoshf',
                                 'sinh': 'csinhf',
                                 'tanh': 'ctanhf',
                                 'acosh': 'cacoshf',
                                 'asinh': 'casinhf',
                                 'atanh': 'catanhf',
                                 'power': 'cpowf',
                                 'exp': 'cexpf',
                                 'ln': 'clogf',
                                 'real': 'crealf',
                                 'imag': 'cimagf',
                                 'conj': 'conjf',
                                 'max_value': 'fmaxf',
                                 'min_value': 'fminf'}}


class UFL2CNodesTranslatorCpp(object):
    """UFL to CNodes translator class."""

    def __init__(self, language, scalar_type="double"):
        self.L = language
        self.force_floats = False
        self.enable_strength_reduction = False
        self.scalar_type = scalar_type

        # Lookup table for handler to call when the "get" method (below) is
        # called, depending on the first argument type.
        self.call_lookup = {ufl.constantvalue.IntValue: self.int_value,
                            ufl.constantvalue.FloatValue: self.float_value,
                            ufl.constantvalue.ComplexValue: self.complex_value,
                            ufl.constantvalue.Zero: self.zero,
                            ufl.algebra.Product: self.product,
                            ufl.algebra.Sum: self.sum,
                            ufl.algebra.Division: self.division,
                            ufl.algebra.Abs: self._cmath,
                            ufl.algebra.Power: self._cmath,
                            ufl.algebra.Real: self._cmath,
                            ufl.algebra.Imag: self._cmath,
                            ufl.algebra.Conj: self._cmath,
                            ufl.classes.GT: self.gt,
                            ufl.classes.GE: self.ge,
                            ufl.classes.EQ: self.eq,
                            ufl.classes.NE: self.ne,
                            ufl.classes.LT: self.lt,
                            ufl.classes.LE: self.le,
                            ufl.classes.AndCondition: self.and_condition,
                            ufl.classes.OrCondition: self.or_condition,
                            ufl.classes.NotCondition: self.not_condition,
                            ufl.classes.Conditional: self.conditional,
                            ufl.classes.MinValue: self._cmath,
                            ufl.classes.MaxValue: self._cmath,
                            ufl.mathfunctions.Sqrt: self._cmath,
                            ufl.mathfunctions.Ln: self._cmath,
                            ufl.mathfunctions.Exp: self._cmath,
                            ufl.mathfunctions.Cos: self._cmath,
                            ufl.mathfunctions.Sin: self._cmath,
                            ufl.mathfunctions.Tan: self._cmath,
                            ufl.mathfunctions.Cosh: self._cmath,
                            ufl.mathfunctions.Sinh: self._cmath,
                            ufl.mathfunctions.Tanh: self._cmath,
                            ufl.mathfunctions.Acos: self._cmath,
                            ufl.mathfunctions.Asin: self._cmath,
                            ufl.mathfunctions.Atan: self._cmath,
                            ufl.mathfunctions.Erf: self._cmath,
                            ufl.mathfunctions.Atan2: self._cmath,
                            ufl.mathfunctions.MathFunction: self.math_function,
                            ufl.mathfunctions.BesselJ: self.bessel_j,
                            ufl.mathfunctions.BesselY: self.bessel_y}

    def get(self, o, *args):
        # Call appropriate handler, depending on the type of o
        otype = type(o)
        if otype in self.call_lookup:
            return self.call_lookup[otype](o, *args)
        else:
            raise RuntimeError(f"Missing C formatting rule for expr type {otype}.")

    def expr(self, o, *args):
        """Raise generic fallback with error message for missing rules."""
        raise RuntimeError(f"Missing C formatting rule for expr type {o._ufl_class_}.")

    # === Formatting rules for scalar literals ===

    def zero(self, o):
        return self.L.LiteralFloat(0.0)

    def float_value(self, o):
        return self.L.LiteralFloat(float(o))

    def int_value(self, o):
        if self.force_floats:
            return self.float_value(o)
        return self.L.LiteralInt(int(o))

    def complex_value(self, o):
        return self.L.LiteralFloat(o.value())

    # === Formatting rules for arithmetic operators ===

    def sum(self, o, a, b):
        return self.L.Add(a, b)

    def product(self, o, a, b):
        return self.L.Mul(a, b)

    def division(self, o, a, b):
        if self.enable_strength_reduction:
            return self.L.Mul(a, self.L.Div(1.0, b))
        else:
            return self.L.Div(a, b)

    # === Formatting rules for conditional expressions ===

    def conditional(self, o, c, t, f):
        return self.L.Conditional(c, t, f)

    def eq(self, o, a, b):
        return self.L.EQ(a, b)

    def ne(self, o, a, b):
        return self.L.NE(a, b)

    def le(self, o, a, b):
        return self.L.LE(a, b)

    def ge(self, o, a, b):
        return self.L.GE(a, b)

    def lt(self, o, a, b):
        return self.L.LT(a, b)

    def gt(self, o, a, b):
        return self.L.GT(a, b)

    def and_condition(self, o, a, b):
        return self.L.And(a, b)

    def or_condition(self, o, a, b):
        return self.L.Or(a, b)

    def not_condition(self, o, a):
        return self.L.Not(a)

    # === Formatting rules for cmath functions ===

    def math_function(self, o, op):
        # Fallback for unhandled MathFunction subclass:
        # attempting to just call it.
        return self.L.Call(o._name, op)

    def _cmath(self, o, *args):
        k = o._ufl_handler_name_
        try:
            name = math_table[self.scalar_type].get(k)
        except Exception as e:
            raise type(e)("Math function not found:", self.scalar_type, k)
        if name is None:
            raise RuntimeError("Not supported in current scalar mode")
        return self.L.Call(name, args)

    # === Formatting rules for bessel functions ===
    # Some Bessel functions exist in gcc, as XSI extensions
    # but not all.

    def bessel_j(self, o, n, v):
        assert "complex" not in self.scalar_type
        n = int(float(n))
        if n == 0:
            return self.L.Call("j0", v)
        elif n == 1:
            return self.L.Call("j1", v)
        else:
            return self.L.Call("jn", (n, v))

    def bessel_y(self, o, n, v):
        assert "complex" not in self.scalar_type
        n = int(float(n))
        if n == 0:
            return self.L.Call("y0", v)
        elif n == 1:
            return self.L.Call("y1", v)
        else:
            return self.L.Call("yn", (n, v))



# Code generation format strings for UFC (Unified Form-assembly Code)
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2018.

declaration = """
extern ufcx_dofmap {factory_name};
"""

factory = """
// Code for dofmap {factory_name}

{sub_dofmaps_initialization}

void tabulate_entity_dofs_{factory_name}(int* restrict dofs, int d, int i)
{{
{tabulate_entity_dofs}
}}

void tabulate_entity_closure_dofs_{factory_name}(int* restrict dofs, int d, int i)
{{
{tabulate_entity_closure_dofs}
}}

{num_entity_dofs_init}

{num_entity_closure_dofs_init}

ufcx_dofmap {factory_name} =
{{
  .signature = {signature},
  .num_global_support_dofs = {num_global_support_dofs},
  .num_element_support_dofs = {num_element_support_dofs},
  .block_size = {block_size},
  .num_entity_dofs = {num_entity_dofs},
  .tabulate_entity_dofs = tabulate_entity_dofs_{factory_name},
  .num_entity_closure_dofs = {num_entity_closure_dofs},
  .tabulate_entity_closure_dofs = tabulate_entity_closure_dofs_{factory_name},
  .num_sub_dofmaps = {num_sub_dofmaps},
  .sub_dofmaps = {sub_dofmaps}
}};

// End of code for dofmap {factory_name}
"""



# Copyright (C) 2004-2019 Garth N. Wells
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import importlib
import io
import logging
import os
import re
import sysconfig
import tempfile
import time
from contextlib import redirect_stdout
from pathlib import Path

import cffi

import ffcx
import ffcx.naming

logger = logging.getLogger("ffcx")

# Get declarations directly from ufcx.h
file_dir = os.path.dirname(os.path.abspath(__file__))
with open(file_dir + "/ufcx.h", "r") as f:
    ufcx_h = ''.join(f.readlines())

header = ufcx_h.split("<HEADER_DECL>")[1].split("</HEADER_DECL>")[0].strip(" /\n")
header = header.replace("{", "{{").replace("}", "}}")
UFC_HEADER_DECL = header + "\n"

UFC_ELEMENT_DECL = '\n'.join(re.findall('typedef struct ufcx_finite_element.*?ufcx_finite_element;', ufcx_h, re.DOTALL))
UFC_DOFMAP_DECL = '\n'.join(re.findall('typedef struct ufcx_dofmap.*?ufcx_dofmap;', ufcx_h, re.DOTALL))
UFC_FORM_DECL = '\n'.join(re.findall('typedef struct ufcx_form.*?ufcx_form;', ufcx_h, re.DOTALL))

UFC_INTEGRAL_DECL = '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_float32\).*?\);', ufcx_h, re.DOTALL))
UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_float64\).*?\);', ufcx_h, re.DOTALL))
UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_complex64\).*?\);', ufcx_h, re.DOTALL))
UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_complex128\).*?\);',
                               ufcx_h, re.DOTALL))
UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_longdouble\).*?\);',
                               ufcx_h, re.DOTALL))

UFC_INTEGRAL_DECL += '\n'.join(re.findall('typedef struct ufcx_integral.*?ufcx_integral;',
                                          ufcx_h, re.DOTALL))
UFC_EXPRESSION_DECL = '\n'.join(re.findall('typedef struct ufcx_expression.*?ufcx_expression;', ufcx_h, re.DOTALL))


def _compute_option_signature(options):
    """Return options signature (some options should not affect signature)."""
    return str(sorted(options.items()))


def get_cached_module(module_name, object_names, cache_dir, timeout):
    """Look for an existing C file and wait for compilation, or if it does not exist, create it."""
    cache_dir = Path(cache_dir)
    c_filename = cache_dir.joinpath(module_name).with_suffix(".c")
    ready_name = c_filename.with_suffix(".c.cached")

    # Ensure cache dir exists
    cache_dir.mkdir(exist_ok=True, parents=True)

    try:
        # Create C file with exclusive access
        open(c_filename, "x")
        return None, None
    except FileExistsError:
        logger.info("Cached C file already exists: " + str(c_filename))
        finder = importlib.machinery.FileFinder(
            str(cache_dir), (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES))
        finder.invalidate_caches()

        # Now, wait for ready
        for i in range(timeout):
            if os.path.exists(ready_name):
                spec = finder.find_spec(module_name)
                if spec is None:
                    raise ModuleNotFoundError("Unable to find JIT module.")
                compiled_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(compiled_module)

                compiled_objects = [getattr(compiled_module.lib, name) for name in object_names]
                return compiled_objects, compiled_module

            logger.info(f"Waiting for {ready_name} to appear.")
            time.sleep(1)
        raise TimeoutError(f"""JIT compilation timed out, probably due to a failed previous compile.
        Try cleaning cache (e.g. remove {c_filename}) or increase timeout option.""")


def _compilation_signature(cffi_extra_compile_args=None, cffi_debug=None):
    """Compute the compilation-inputs part of the signature.

    Used to avoid cache conflicts across Python versions, architectures, installs.

    - SOABI includes platform, Python version, debug flags
    - CFLAGS includes prefixes, arch targets
    """
    return (
        str(cffi_extra_compile_args)
        + str(cffi_debug)
        + sysconfig.get_config_var("CFLAGS")
        + sysconfig.get_config_var("SOABI")
    )


def compile_elements(elements, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
                     cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
    """Compile a list of UFL elements and dofmaps into Python objects."""
    p = ffcx.options.get_options(options)

    # Get a signature for these elements
    module_name = 'libffcx_elements_' + \
        ffcx.naming.compute_signature(elements, _compute_option_signature(p)
                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))

    names = []
    for e in elements:
        name = ffcx.naming.finite_element_name(e, module_name)
        names.append(name)
        name = ffcx.naming.dofmap_name(e, module_name)
        names.append(name)

    if cache_dir is not None:
        cache_dir = Path(cache_dir)
        obj, mod = get_cached_module(module_name, names, cache_dir, timeout)
        if obj is not None:
            # Pair up elements with dofmaps
            obj = list(zip(obj[::2], obj[1::2]))
            return obj, mod, (None, None)
    else:
        cache_dir = Path(tempfile.mkdtemp())

    try:
        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL
        element_template = "extern ufcx_finite_element {name};\n"
        dofmap_template = "extern ufcx_dofmap {name};\n"
        for i in range(len(elements)):
            decl += element_template.format(name=names[i * 2])
            decl += dofmap_template.format(name=names[i * 2 + 1])

        impl = _compile_objects(decl, elements, names, module_name, p, cache_dir,
                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
    except Exception:
        # remove c file so that it will not timeout next time
        c_filename = cache_dir.joinpath(module_name + ".c")
        os.replace(c_filename, c_filename.with_suffix(".c.failed"))
        raise

    objects, module = _load_objects(cache_dir, module_name, names)
    # Pair up elements with dofmaps
    objects = list(zip(objects[::2], objects[1::2]))
    return objects, module, (decl, impl)


def compile_forms(forms, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
                  cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
    """Compile a list of UFL forms into UFC Python objects."""
    p = ffcx.options.get_options(options)

    # Get a signature for these forms
    module_name = 'libffcx_forms_' + \
        ffcx.naming.compute_signature(forms, _compute_option_signature(p)
                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))

    form_names = [ffcx.naming.form_name(form, i, module_name) for i, form in enumerate(forms)]

    if cache_dir is not None:
        cache_dir = Path(cache_dir)
        obj, mod = get_cached_module(module_name, form_names, cache_dir, timeout)
        if obj is not None:
            return obj, mod, (None, None)
    else:
        cache_dir = Path(tempfile.mkdtemp())

    try:
        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL + \
            UFC_INTEGRAL_DECL + UFC_FORM_DECL

        form_template = "extern ufcx_form {name};\n"
        for name in form_names:
            decl += form_template.format(name=name)

        impl = _compile_objects(decl, forms, form_names, module_name, p, cache_dir,
                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
    except Exception:
        # remove c file so that it will not timeout next time
        c_filename = cache_dir.joinpath(module_name + ".c")
        os.replace(c_filename, c_filename.with_suffix(".c.failed"))
        raise

    obj, module = _load_objects(cache_dir, module_name, form_names)
    return obj, module, (decl, impl)


def compile_expressions(expressions, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
                        cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
    """Compile a list of UFL expressions into UFC Python objects.

    Options
    ----------
    expressions
        List of (UFL expression, evaluation points).

    """
    p = ffcx.options.get_options(options)

    module_name = 'libffcx_expressions_' + \
        ffcx.naming.compute_signature(expressions, _compute_option_signature(p)
                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))
    expr_names = [ffcx.naming.expression_name(expression, module_name) for expression in expressions]

    if cache_dir is not None:
        cache_dir = Path(cache_dir)
        obj, mod = get_cached_module(module_name, expr_names, cache_dir, timeout)
        if obj is not None:
            return obj, mod, (None, None)
    else:
        cache_dir = Path(tempfile.mkdtemp())

    try:
        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL + \
            UFC_INTEGRAL_DECL + UFC_FORM_DECL + UFC_EXPRESSION_DECL

        expression_template = "extern ufcx_expression {name};\n"
        for name in expr_names:
            decl += expression_template.format(name=name)

        impl = _compile_objects(decl, expressions, expr_names, module_name, p, cache_dir,
                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
    except Exception:
        # remove c file so that it will not timeout next time
        c_filename = cache_dir.joinpath(module_name + ".c")
        os.replace(c_filename, c_filename.with_suffix(".c.failed"))
        raise

    obj, module = _load_objects(cache_dir, module_name, expr_names)
    return obj, module, (decl, impl)


def _compile_objects(decl, ufl_objects, object_names, module_name, options, cache_dir,
                     cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries):

    import ffcx.compiler

    # JIT uses module_name as prefix, which is needed to make names of all struct/function
    # unique across modules
    _, code_body = ffcx.compiler.compile_ufl_objects(ufl_objects, prefix=module_name, options=options)

    ffibuilder = cffi.FFI()
    ffibuilder.set_source(module_name, code_body, include_dirs=[ffcx.codegeneration.get_include_path()],
                          extra_compile_args=cffi_extra_compile_args, libraries=cffi_libraries)
    ffibuilder.cdef(decl)

    c_filename = cache_dir.joinpath(module_name + ".c")
    ready_name = c_filename.with_suffix(".c.cached")

    # Compile (ensuring that compile dir exists)
    cache_dir.mkdir(exist_ok=True, parents=True)

    logger.info(79 * "#")
    logger.info("Calling JIT C compiler")
    logger.info(79 * "#")

    t0 = time.time()
    f = io.StringIO()
    with redirect_stdout(f):
        ffibuilder.compile(tmpdir=cache_dir, verbose=True, debug=cffi_debug)
    s = f.getvalue()
    if (cffi_verbose):
        print(s)

    logger.info("JIT C compiler finished in {:.4f}".format(time.time() - t0))

    # Create a "status ready" file. If this fails, it is an error,
    # because it should not exist yet.
    # Copy the stdout verbose output of the build into the ready file
    fd = open(ready_name, "x")
    fd.write(s)
    fd.close()

    return code_body


def _load_objects(cache_dir, module_name, object_names):

    # Create module finder that searches the compile path
    finder = importlib.machinery.FileFinder(
        str(cache_dir), (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES))

    # Find module. Clear search cache to be sure dynamically created
    # (new) modules are found
    finder.invalidate_caches()
    spec = finder.find_spec(module_name)
    if spec is None:
        raise ModuleNotFoundError("Unable to find JIT module.")

    # Load module
    compiled_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(compiled_module)

    compiled_objects = []
    for name in object_names:
        obj = getattr(compiled_module.lib, name)
        compiled_objects.append(obj)

    return compiled_objects, compiled_module



# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# Test all algebra operators on Coefficients.
from ufl import (Coefficient, FiniteElement, acos, asin, atan, bessel_J,
                 bessel_Y, cos, dx, erf, exp, ln, sin, sqrt, tan, triangle)

element = FiniteElement("Lagrange", triangle, 1)

c0 = Coefficient(element)
c1 = Coefficient(element)

s0 = 3 * c0 - c1
p0 = c0 * c1
f0 = c0 / c1

integrand = sqrt(c0) + sqrt(s0) + sqrt(p0) + sqrt(f0)\
    + exp(c0) + exp(s0) + exp(p0) + exp(f0)\
    + ln(c0) + ln(s0) + ln(p0) + ln(f0)\
    + cos(c0) + cos(s0) + cos(p0) + cos(f0)\
    + sin(c0) + sin(s0) + sin(p0) + sin(f0)\
    + tan(c0) + tan(s0) + tan(p0) + tan(f0)\
    + acos(c0) + acos(s0) + acos(p0) + acos(f0)\
    + asin(c0) + asin(s0) + asin(p0) + asin(f0)\
    + atan(c0) + atan(s0) + atan(p0) + atan(f0)\
    + erf(c0) + erf(s0) + erf(p0) + erf(f0)\
    + bessel_J(1, c0) + bessel_J(1, s0) + bessel_J(0, p0) + bessel_J(0, f0)\
    + bessel_Y(1, c0) + bessel_Y(1, s0) + bessel_Y(0, p0) + bessel_Y(0, f0)

a = integrand * dx



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Linearized data structure for the computational graph."""

import logging

import numpy

import ufl
from ffcx.ir.analysis.modified_terminals import is_modified_terminal
from ffcx.ir.analysis.reconstruct import reconstruct
from ffcx.ir.analysis.valuenumbering import ValueNumberer

logger = logging.getLogger("ffcx")


class ExpressionGraph(object):
    """A directed multi-edge graph.

    ExpressionGraph allows multiple edges between the same nodes,
    and respects the insertion order of nodes and edges.
    """

    def __init__(self):

        # Data structures for directed multi-edge graph
        self.nodes = {}
        self.out_edges = {}
        self.in_edges = {}

    def number_of_nodes(self):
        return len(self.nodes)

    def add_node(self, key, **kwargs):
        """Add a node with optional properties."""
        self.nodes[key] = kwargs
        self.out_edges[key] = []
        self.in_edges[key] = []

    def add_edge(self, node1, node2):
        """Add a directed edge from node1 to node2."""
        if node1 not in self.nodes or node2 not in self.nodes:
            raise KeyError("Adding edge to unknown node")

        self.out_edges[node1] += [node2]
        self.in_edges[node2] += [node1]


def build_graph_vertices(expressions, skip_terminal_modifiers=False):
    # Count unique expression nodes
    G = ExpressionGraph()
    G.e2i = _count_nodes_with_unique_post_traversal(expressions, skip_terminal_modifiers)

    # Invert the map to get index->expression
    GV = sorted(G.e2i, key=G.e2i.get)

    # Add nodes to 'new' graph structure
    for i, v in enumerate(GV):
        G.add_node(i, expression=v)

    for comp, expr in enumerate(expressions):
        # Get vertex index representing input expression root
        V_target = G.e2i[expr]
        G.nodes[V_target]['target'] = True
        G.nodes[V_target]['component'] = G.nodes[V_target].get("component", [])
        G.nodes[V_target]['component'].append(comp)

    return G


def build_scalar_graph(expression):
    """Build list representation of expression graph covering the given expressions."""
    # Populate with vertices
    G = build_graph_vertices([expression], skip_terminal_modifiers=False)

    # Build more fine grained computational graph of scalar subexpressions
    scalar_expressions = rebuild_with_scalar_subexpressions(G)

    # Build new list representation of graph where all
    # vertices of V represent single scalar operations
    G = build_graph_vertices(scalar_expressions, skip_terminal_modifiers=True)

    # Compute graph edges
    V_deps = []
    for i, v in G.nodes.items():
        expr = v['expression']
        if expr._ufl_is_terminal_ or expr._ufl_is_terminal_modifier_:
            V_deps.append(())
        else:
            V_deps.append([G.e2i[o] for o in expr.ufl_operands])

    for i, edges in enumerate(V_deps):
        for j in edges:
            if i == j:
                continue
            G.add_edge(i, j)

    return G


def rebuild_with_scalar_subexpressions(G):
    """Build a new expression2index mapping where each subexpression is scalar valued.

    Input:
    - G.e2i
    - G.V
    - G.V_symbols
    - G.total_unique_symbols

    Output:
    - NV   - Array with reverse mapping from index to expression
    - nvs  - Tuple of ne2i indices corresponding to the last vertex of G.V
    """
    # Compute symbols over graph and rebuild scalar expression
    #
    # New expression which represents usually an algebraic operation
    # generates a new symbol
    value_numberer = ValueNumberer(G)

    # V_symbols maps an index of a node to a list of
    # symbols which are present in that node
    V_symbols = value_numberer.compute_symbols()
    total_unique_symbols = value_numberer.symbol_count

    # Array to store the scalar subexpression in for each symbol
    W = numpy.empty(total_unique_symbols, dtype=object)

    # Iterate over each graph node in order
    for i, v in G.nodes.items():
        expr = v['expression']
        # Find symbols of v components
        vs = V_symbols[i]

        # Skip if there's nothing new here (should be the case for indexing types)
        # New symbols are not given to indexing types, so W[symbol] already equals
        # an expression, since it was assigned to the symbol in a previous loop
        # cycle
        if all(W[s] is not None for s in vs):
            continue

        if is_modified_terminal(expr):
            sh = expr.ufl_shape
            if sh:
                # Store each terminal expression component. We may not
                # actually need all of these later, but that will be
                # optimized away.
                # Note: symmetries will be dealt with in the value numbering.
                ws = [expr[c] for c in ufl.permutation.compute_indices(sh)]
            else:
                # Store single modified terminal expression component
                if len(vs) != 1:
                    raise RuntimeError("Expecting single symbol for scalar valued modified terminal.")
                ws = [expr]
            # FIXME: Replace ws[:] with 0's if its table is empty
            # Possible redesign: loop over modified terminals only first,
            # then build tables for them, set W[s] = 0.0 for modified terminals with zero table,
            # then loop over non-(modified terminal)s to reconstruct expression.
        else:
            # Find symbols of operands
            sops = []
            for j, vop in enumerate(expr.ufl_operands):
                if isinstance(vop, ufl.classes.MultiIndex):
                    # TODO: Store MultiIndex in G.V and allocate a symbol to it for this to work
                    if not isinstance(expr, ufl.classes.IndexSum):
                        raise RuntimeError("Not expecting a %s." % type(expr))
                    sops.append(())
                else:
                    # TODO: Build edge datastructure and use instead?
                    # k = G.E[i][j]
                    k = G.e2i[vop]
                    sops.append(V_symbols[k])

            # Fetch reconstructed operand expressions
            wops = [tuple(W[k] for k in so) for so in sops]

            # Reconstruct scalar subexpressions of v
            ws = reconstruct(expr, wops)

            # Store all scalar subexpressions for v symbols
            if len(vs) != len(ws):
                raise RuntimeError("Expecting one symbol for each expression.")

        # Store each new scalar subexpression in W at the index of its symbol
        handled = set()
        for s, w in zip(vs, ws):
            if W[s] is None:
                W[s] = w
                handled.add(s)
            else:
                assert s in handled  # Result of symmetry! - but I think this never gets reached anyway (CNR)

    # Find symbols of final v from input graph
    vs = V_symbols[-1]
    scalar_expressions = W[vs]
    return scalar_expressions


def _count_nodes_with_unique_post_traversal(expressions, skip_terminal_modifiers=False):
    """Yield o for each node o in expr, child before parent.

    Never visits a node twice.
    """

    def getops(e):
        """Get a modifiable list of operands of e, optionally treating modified terminals as a unit."""
        # TODO: Maybe use e._ufl_is_terminal_modifier_
        if e._ufl_is_terminal_ or (skip_terminal_modifiers and is_modified_terminal(e)):
            return []
        else:
            return list(e.ufl_operands)

    e2i = {}
    stack = [(expr, getops(expr)) for expr in reversed(expressions)]
    while stack:
        expr, ops = stack[-1]

        if expr in e2i:
            stack.pop()
            continue

        for i, o in enumerate(ops):
            if o is not None and o not in e2i:
                stack.append((o, getops(o)))
                ops[i] = None
                break
        else:
            if not isinstance(expr, (ufl.classes.MultiIndex, ufl.classes.Label)):
                count = len(e2i)
                e2i[expr] = count
            stack.pop()
    return e2i



# Copyright (C) 2009-2018 FEniCS Project
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

"""FEniCS Form Compiler (FFCx).

FFCx compiles finite element variational forms into C code.

"""

import logging

import pkg_resources

# Import default options
from ffcx.options import get_options  # noqa: F401

__version__ = pkg_resources.get_distribution("fenics-ffcx").version

logging.basicConfig()
logger = logging.getLogger("ffcx")
logging.captureWarnings(capture=True)



# Copyright (C) 2021 Matthew Scroggs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy

import basix


def write_table(L, tablename, cellname, type: str):
    if tablename == "facet_edge_vertices":
        return facet_edge_vertices(L, tablename, cellname)
    if tablename == "reference_facet_jacobian":
        return reference_facet_jacobian(L, tablename, cellname, type)
    if tablename == "reference_cell_volume":
        return reference_cell_volume(L, tablename, cellname, type)
    if tablename == "reference_facet_volume":
        return reference_facet_volume(L, tablename, cellname, type)
    if tablename == "reference_edge_vectors":
        return reference_edge_vectors(L, tablename, cellname, type)
    if tablename == "facet_reference_edge_vectors":
        return facet_reference_edge_vectors(L, tablename, cellname, type)
    if tablename == "reference_facet_normals":
        return reference_facet_normals(L, tablename, cellname, type)
    if tablename == "facet_orientation":
        return facet_orientation(L, tablename, cellname, type)
    raise ValueError(f"Unknown geometry table name: {tablename}")


def facet_edge_vertices(L, tablename, cellname):
    celltype = getattr(basix.CellType, cellname)
    topology = basix.topology(celltype)
    triangle_edges = basix.topology(basix.CellType.triangle)[1]
    quadrilateral_edges = basix.topology(basix.CellType.quadrilateral)[1]

    if len(topology) != 4:
        raise ValueError("Can only get facet edges for 3D cells.")

    edge_vertices = []
    for facet in topology[-2]:
        if len(facet) == 3:
            edge_vertices += [[facet[i] for i in edge] for edge in triangle_edges]
        elif len(facet) == 4:
            edge_vertices += [[facet[i] for i in edge] for edge in quadrilateral_edges]
        else:
            raise ValueError("Only triangular and quadrilateral faces supported.")

    out = numpy.array(edge_vertices, dtype=int)
    return L.ArrayDecl("static const unsigned int", f"{cellname}_{tablename}", out.shape, out)


def reference_facet_jacobian(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    out = basix.cell.facet_jacobians(celltype)
    return L.ArrayDecl(f"static const {type}", f"{cellname}_{tablename}", out.shape, out)


def reference_cell_volume(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    out = basix.cell.volume(celltype)
    return L.VariableDecl(f"static const {type}", f"{cellname}_{tablename}", out)


def reference_facet_volume(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    volumes = basix.cell.facet_reference_volumes(celltype)
    for i in volumes[1:]:
        if not numpy.isclose(i, volumes[0]):
            raise ValueError("Reference facet volume not supported for this cell type.")
    return L.VariableDecl(f"static const {type}", f"{cellname}_{tablename}", volumes[0])


def reference_edge_vectors(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    topology = basix.topology(celltype)
    geometry = basix.geometry(celltype)
    edge_vectors = [geometry[j] - geometry[i] for i, j in topology[1]]
    out = numpy.array(edge_vectors[cellname])
    return L.ArrayDecl(f"static const {type}", f"{cellname}_{tablename}", out.shape, out)


def facet_reference_edge_vectors(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    topology = basix.topology(celltype)
    geometry = basix.geometry(celltype)
    triangle_edges = basix.topology(basix.CellType.triangle)[1]
    quadrilateral_edges = basix.topology(basix.CellType.quadrilateral)[1]

    if len(topology) != 4:
        raise ValueError("Can only get facet edges for 3D cells.")

    edge_vectors = []
    for facet in topology[-2]:
        if len(facet) == 3:
            edge_vectors += [geometry[facet[j]] - geometry[facet[i]] for i, j in triangle_edges]
        elif len(facet) == 4:
            edge_vectors += [geometry[facet[j]] - geometry[facet[i]] for i, j in quadrilateral_edges]
        else:
            raise ValueError("Only triangular and quadrilateral faces supported.")

    out = numpy.array(edge_vectors)
    return L.ArrayDecl(f"static const {type}", f"{cellname}_{tablename}", out.shape, out)


def reference_facet_normals(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    out = basix.cell.facet_outward_normals(celltype)
    return L.ArrayDecl(f"static const {type}", f"{cellname}_{tablename}", out.shape, out)


def facet_orientation(L, tablename, cellname, type: str):
    celltype = getattr(basix.CellType, cellname)
    out = basix.cell.facet_orientations(celltype)
    return L.ArrayDecl(f"static const {type}", f"{cellname}_{tablename}", out.shape, out)



# Copyright (C) 2019 Michal Habera
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

declaration = """
extern ufcx_expression {factory_name};

// Helper used to create expression using name which was given to the
// expression in the UFL file.
// This helper is called in user c++ code.
//
extern ufcx_expression* {name_from_uflfile};
"""

factory = """
// Code for expression {factory_name}

void tabulate_tensor_{factory_name}({scalar_type}* restrict A,
                                    const {scalar_type}* restrict w,
                                    const {scalar_type}* restrict c,
                                    const {geom_type}* restrict coordinate_dofs,
                                    const int* restrict entity_local_index,
                                    const uint8_t* restrict quadrature_permutation)
{{
{tabulate_expression}
}}

{points_init}
{value_shape_init}
{original_coefficient_positions_init}
{function_spaces_alloc}
{function_spaces_init}
{coefficient_names_init}
{constant_names_init}


ufcx_expression {factory_name} =
{{
  .tabulate_tensor_{np_scalar_type} = tabulate_tensor_{factory_name},
  .num_coefficients = {num_coefficients},
  .num_constants = {num_constants},
  .original_coefficient_positions = {original_coefficient_positions},
  .coefficient_names = {coefficient_names},
  .constant_names = {constant_names},
  .num_points = {num_points},
  .topological_dimension = {topological_dimension},
  .points = {points},
  .value_shape = {value_shape},
  .num_components = {num_components},
  .rank = {rank},
  .function_spaces = {function_spaces}
}};

// Alias name
ufcx_expression* {name_from_uflfile} = &{factory_name};

// End of code for expression {factory_name}
"""



# Copyright (C) 2019 Chris Richardson
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy as np
import pytest

import ffcx.codegeneration.jit
import ufl
from ffcx.naming import cdtype_to_numpy, scalar_to_value_type


@pytest.mark.parametrize("mode",
                         [
                             "double",
                             "float",
                             "long double",
                             "double _Complex",
                             "float _Complex"
                         ])
def test_additive_facet_integral(mode, compile_args):
    cell = ufl.triangle
    element = ufl.FiniteElement("Lagrange", cell, 1)
    u, v = ufl.TrialFunction(element), ufl.TestFunction(element)
    a = ufl.inner(u, v) * ufl.ds
    forms = [a]
    compiled_forms, module, code = ffcx.codegeneration.jit.compile_forms(
        forms, options={'scalar_type': mode}, cffi_extra_compile_args=compile_args)

    for f, compiled_f in zip(forms, compiled_forms):
        assert compiled_f.rank == len(f.arguments())

    ffi = module.ffi
    form0 = compiled_forms[0]

    assert form0.num_integrals(module.lib.exterior_facet) == 1
    ids = form0.integral_ids(module.lib.exterior_facet)
    assert ids[0] == -1

    default_integral = form0.integrals(module.lib.exterior_facet)[0]

    np_type = cdtype_to_numpy(mode)
    A = np.zeros((3, 3), dtype=np_type)
    w = np.array([], dtype=np_type)
    c = np.array([], dtype=np_type)
    facets = np.array([0], dtype=np.int32)
    perm = np.array([0], dtype=np.uint8)

    geom_type = scalar_to_value_type(mode)
    np_gtype = cdtype_to_numpy(geom_type)
    coords = np.array([0.0, 2.0, 0.0,
                       np.sqrt(3.0), -1.0, 0.0,
                       -np.sqrt(3.0), -1.0, 0.0], dtype=np_gtype)

    kernel = getattr(default_integral, f"tabulate_tensor_{np_type}")

    for i in range(3):
        facets[0] = i
        kernel(ffi.cast('{type} *'.format(type=mode), A.ctypes.data),
               ffi.cast('{type} *'.format(type=mode), w.ctypes.data),
               ffi.cast('{type} *'.format(type=mode), c.ctypes.data),
               ffi.cast(f'{geom_type} *', coords.ctypes.data),
               ffi.cast('int *', facets.ctypes.data),
               ffi.cast('uint8_t *', perm.ctypes.data))

        assert np.isclose(A.sum(), np.sqrt(12) * (i + 1))


@pytest.mark.parametrize("mode", ["double", "float", "long double", "double _Complex", "float _Complex"])
def test_additive_cell_integral(mode, compile_args):
    cell = ufl.triangle
    element = ufl.FiniteElement("Lagrange", cell, 1)
    u, v = ufl.TrialFunction(element), ufl.TestFunction(element)
    a = ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx
    forms = [a]
    compiled_forms, module, code = ffcx.codegeneration.jit.compile_forms(
        forms, options={'scalar_type': mode}, cffi_extra_compile_args=compile_args)

    for f, compiled_f in zip(forms, compiled_forms):
        assert compiled_f.rank == len(f.arguments())

    ffi = module.ffi
    form0 = compiled_forms[0]

    assert form0.num_integrals(module.lib.cell) == 1
    ids = form0.integral_ids(module.lib.cell)
    assert ids[0] == -1

    default_integral = form0.integrals(0)[0]

    np_type = cdtype_to_numpy(mode)
    A = np.zeros((3, 3), dtype=np_type)
    w = np.array([], dtype=np_type)
    c = np.array([], dtype=np_type)

    geom_type = scalar_to_value_type(mode)
    np_gtype = cdtype_to_numpy(geom_type)
    coords = np.array([0.0, 2.0, 0.0,
                       np.sqrt(3.0), -1.0, 0.0,
                       -np.sqrt(3.0), -1.0, 0.0], dtype=np_gtype)

    kernel = getattr(default_integral, f"tabulate_tensor_{np_type}")

    kernel(ffi.cast('{type} *'.format(type=mode), A.ctypes.data),
           ffi.cast('{type} *'.format(type=mode), w.ctypes.data),
           ffi.cast('{type} *'.format(type=mode), c.ctypes.data),
           ffi.cast(f'{geom_type} *', coords.ctypes.data), ffi.NULL, ffi.NULL)

    A0 = np.array(A)
    for i in range(3):
        kernel(ffi.cast('{type} *'.format(type=mode), A.ctypes.data),
               ffi.cast('{type} *'.format(type=mode), w.ctypes.data),
               ffi.cast('{type} *'.format(type=mode), c.ctypes.data),
               ffi.cast(f'{geom_type} *', coords.ctypes.data), ffi.NULL, ffi.NULL)

        assert np.all(np.isclose(A, (i + 2) * A0))



# Copyright (C) 2016 Lizao Li
#
# The bilinear form a(u, v) and linear form L(v) for
# Biharmonic equation in Hellan-Herrmann-Johnson (HHJ)
# formulation.
from ufl import (Coefficient, FacetNormal, FiniteElement, TestFunctions,
                 TrialFunctions, dot, dS, ds, dx, grad, inner, jump, triangle)

HHJ = FiniteElement('HHJ', triangle, 2)
CG = FiniteElement('CG', triangle, 3)
mixed_element = HHJ * CG

(sigma, u) = TrialFunctions(mixed_element)
(tau, v) = TestFunctions(mixed_element)
f = Coefficient(CG)


def b(sigma, v):
    n = FacetNormal(triangle)
    return inner(sigma, grad(grad(v))) * dx \
        - dot(dot(sigma('+'), n('+')), n('+')) * jump(grad(v), n) * dS \
        - dot(dot(sigma, n), n) * dot(grad(v), n) * ds


a = inner(sigma, tau) * dx - b(tau, u) + b(sigma, v)
L = f * v * dx
